from keras_cv_attention_models.efficientnet.efficientnet_v2 import (
    EfficientNetV2,
    EfficientNetV2B0,
    EfficientNetV2B1,
    EfficientNetV2B2,
    EfficientNetV2B3,
    EfficientNetV2T,
    EfficientNetV2T_GC,
    EfficientNetV2S,
    EfficientNetV2M,
    EfficientNetV2L,
    EfficientNetV2XL,
)
from keras_cv_attention_models.efficientnet.efficientnet_v1 import (
    EfficientNetV1,
    EfficientNetV1B0,
    EfficientNetV1B1,
    EfficientNetV1B2,
    EfficientNetV1B3,
    EfficientNetV1B4,
    EfficientNetV1B5,
    EfficientNetV1B6,
    EfficientNetV1B7,
    EfficientNetV1L2,
    EfficientNetV1Lite0,
    EfficientNetV1Lite1,
    EfficientNetV1Lite2,
    EfficientNetV1Lite3,
    EfficientNetV1Lite4,
)
from keras_cv_attention_models.efficientnet.efficientnet_edgetpu import (
    EfficientNetEdgeTPU,
    EfficientNetEdgeTPUSmall,
    EfficientNetEdgeTPUMedium,
    EfficientNetEdgeTPULarge,
)

__v2_head_doc__ = """
Keras implementation of [Official efficientnetv2](https://github.com/google/automl/tree/master/efficientnetv2).
Paper [arXiv 2104.00298 EfficientNetV2: Smaller Models and Faster Training](https://arxiv.org/abs/2104.00298) by Mingxing Tan, Quoc V. Le.
"""

__tail_doc__ = """  input_shape: it should have exactly 3 inputs channels. Set `(None, None, 3)` for dynamic.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  dropout: dropout rate if top layers is included.
  first_strides: is used in the first Conv2D layer.
  drop_connect_rate: is used for: Deep Networks with Stochastic Depth.
      arxiv: https://arxiv.org/abs/1603.09382.
      If not `0`, will add a `Dropout` layer for each deep branch changes
      from `0 --> drop_connect_rate` for `top --> bottom` layers. Or `0` to disable.
      Default `0` for `EfficientNetV2`, `0.2` for `EfficientNetV1`.
  classifier_activation: A `str` or callable.
      The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  include_preprocessing: Boolean value if add preprocessing `Rescale` / `Normalization` after `Input`, will expect input value in range `[0, 255]`.
      Note that different preprocessing will applied for different `rescale_mode`. It depends on `pretained` and `model_type`.
      - For all V1 models, `rescale_mode`s are "torch".
      - For "21k" pretrained V2 models, `rescale_mode`s are all "tf"
      - For "imagenet" pretrained V2 models, "bx" models are all "torch", ["s", "m", "l", "xl"] are "tf".
      Default `False`.
  pretrained: value in {pretrained}.
      Will try to download and load pre-trained model weights if not None.

Returns:
    A `keras.Model` instance.
"""

EfficientNetV2.__doc__ = __v2_head_doc__ + """
Args:
  model_type: is the pre-defined model, value in ["t", "s", "m", "l", "xl", "b0", "b1", "b2", "b3"].
  model_name: string, model name.
""" + __tail_doc__.format(pretrained=[None, "imagenet", "imagenet21k", "imagenet21k-ft1k"]) + """
Model architectures:
  | V2 Model                   | Params | FLOPs  | Input | Top1 Acc |
  | -------------------------- | ------ | ------ | ----- | -------- |
  | EfficientNetV2B0           | 7.1M   | 0.72G  | 224   | 78.7     |
  | - ImageNet21k-ft1k         | 7.1M   | 0.72G  | 224   | 77.55?   |
  | EfficientNetV2B1           | 8.1M   | 1.21G  | 240   | 79.8     |
  | - ImageNet21k-ft1k         | 8.1M   | 1.21G  | 240   | 79.03?   |
  | EfficientNetV2B2           | 10.1M  | 1.71G  | 260   | 80.5     |
  | - ImageNet21k-ft1k         | 10.1M  | 1.71G  | 260   | 79.48?   |
  | EfficientNetV2B3           | 14.4M  | 3.03G  | 300   | 82.1     |
  | - ImageNet21k-ft1k         | 14.4M  | 3.03G  | 300   | 82.46?   |
  | EfficientNetV2T            | 13.6M  | 3.18G  | 288   | 82.34    |
  | EfficientNetV2T_GC         | 13.7M  | 3.19G  | 288   | 82.46    |
  | EfficientNetV2S            | 21.5M  | 8.41G  | 384   | 83.9     |
  | - ImageNet21k-ft1k         | 21.5M  | 8.41G  | 384   | 84.9     |
  | EfficientNetV2M            | 54.1M  | 24.69G | 480   | 85.2     |
  | - ImageNet21k-ft1k         | 54.1M  | 24.69G | 480   | 86.2     |
  | EfficientNetV2L            | 119.5M | 56.27G | 480   | 85.7     |
  | - ImageNet21k-ft1k         | 119.5M | 56.27G | 480   | 86.9     |
  | EfficientNetV2XL, 21k-ft1k | 206.8M | 93.66G | 512   | 87.2     |

Training configures: `Eval size` is used as the default model `input_shape` for each model type.
  | Model            | Train size | Eval size | Dropout | Randaug | Mixup |
  | ---------------- | ---------- | --------- | ------- | ------- | ----- |
  | EfficientNetV2B0 | 192        | 224       | 0.2     | 0       | 0     |
  | EfficientNetV2B1 | 192        | 240       | 0.2     | 0       | 0     |
  | EfficientNetV2B2 | 208        | 260       | 0.3     | 0       | 0     |
  | EfficientNetV2B3 | 240        | 300       | 0.3     | 0       | 0     |
  | EfficientNetV2T  | 224        | 320       | 0.2     | 0       | 0     |
  | EfficientNetV2S  | 300        | 384       | 0.2     | 10      | 0     |
  | EfficientNetV2M  | 384        | 480       | 0.3     | 15      | 0.2   |
  | EfficientNetV2L  | 384        | 480       | 0.4     | 20      | 0.5   |
  | EfficientNetV2XL | 384        | 512       | 0.4     | 20      | 0.5   |
"""

__v2_default_doc__ = __v2_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained=[None, "imagenet", "imagenet21k", "imagenet21k-ft1k"]) + """
Training configures: `Eval size` is used as the default model `input_shape`.
  | Model   | Train size | Eval size | Dropout | Randaug | Mixup |
  | ------- | ---------- | --------- | ------- | ------- | ----- |
  {train_config}
"""

EfficientNetV2B0.__doc__ = __v2_default_doc__.format(train_config="| EffV2B0 | 192        | 224       | 0.2     | 0       | 0     |")
EfficientNetV2B1.__doc__ = __v2_default_doc__.format(train_config="| EffV2B1 | 192        | 240       | 0.2     | 0       | 0     |")
EfficientNetV2B2.__doc__ = __v2_default_doc__.format(train_config="| EffV2B2 | 208        | 260       | 0.3     | 0       | 0     |")
EfficientNetV2B3.__doc__ = __v2_default_doc__.format(train_config="| EffV2B3 | 240        | 300       | 0.3     | 0       | 0     |")
EfficientNetV2S.__doc__ = __v2_default_doc__.format(train_config="| EffV2S  | 300        | 384       | 0.2     | 10      | 0     |")
EfficientNetV2M.__doc__ = __v2_default_doc__.format(train_config="| EffV2M  | 384        | 480       | 0.3     | 15      | 0.2   |")
EfficientNetV2L.__doc__ = __v2_default_doc__.format(train_config="| EffV2L  | 384        | 480       | 0.4     | 20      | 0.5   |")
EfficientNetV2XL.__doc__ = __v2_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained=[None, "imagenet21k", "imagenet21k-ft1k"]) + """
Training configures: `Eval size` is used as the default model `input_shape`.
  | Model   | Train size | Eval size | Dropout | Randaug | Mixup |
  | ------- | ---------- | --------- | ------- | ------- | ----- |
  | EffV2XL | 384        | 512       | 0.4     | 20      | 0.5   |
"""
EfficientNetV2T.__doc__ = __v2_head_doc__ + """Architecture and weights from [Github rwightman/pytorch-image-models](https://github.com/rwightman/pytorch-image-models#july-5-9-2021).

Args:
""" + __tail_doc__.format(pretrained=[None, "imagenet"])
EfficientNetV2T_GC.__doc__ = EfficientNetV2T.__doc__

__v1_head_doc__ = """
Keras implementation of [Github tensorflow/tpu/efficientnet](https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet).
Paper [PDF 1911.04252 Self-training with Noisy Student improves ImageNet classification](https://arxiv.org/pdf/1911.04252.pdf).
"""

EfficientNetV1.__doc__ = __v1_head_doc__ + """
Args:
  model_name: string, model name.
""" + __tail_doc__.format(pretrained=[None, "imagenet", "noisy_student"]) + """
Model architectures:
  | V1 Model                       | Params | FLOPs   | Input | Top1 Acc |
  | ------------------------------ | ------ | ------- | ----- | -------- |
  | EfficientNetV1B0               | 5.3M   | 0.39G   | 224   | 77.6     |
  | - NoisyStudent                 | 5.3M   | 0.39G   | 224   | 78.8     |
  | EfficientNetV1B1               | 7.8M   | 0.70G   | 240   | 79.6     |
  | - NoisyStudent                 | 7.8M   | 0.70G   | 240   | 81.5     |
  | EfficientNetV1B2               | 9.1M   | 1.01G   | 260   | 80.5     |
  | - NoisyStudent                 | 9.1M   | 1.01G   | 260   | 82.4     |
  | EfficientNetV1B3               | 12.2M  | 1.86G   | 300   | 81.9     |
  | - NoisyStudent                 | 12.2M  | 1.86G   | 300   | 84.1     |
  | EfficientNetV1B4               | 19.3M  | 4.46G   | 380   | 83.3     |
  | - NoisyStudent                 | 19.3M  | 4.46G   | 380   | 85.3     |
  | EfficientNetV1B5               | 30.4M  | 10.40G  | 456   | 84.3     |
  | - NoisyStudent                 | 30.4M  | 10.40G  | 456   | 86.1     |
  | EfficientNetV1B6               | 43.0M  | 19.29G  | 528   | 84.8     |
  | - NoisyStudent                 | 43.0M  | 19.29G  | 528   | 86.4     |
  | EfficientNetV1B7               | 66.3M  | 38.13G  | 600   | 85.2     |
  | - NoisyStudent                 | 66.3M  | 38.13G  | 600   | 86.9     |
  | EfficientNetV1L2, NoisyStudent | 480.3M | 477.98G | 800   | 88.4     |


Training configures:
  | Model            | Input resolution | Dropout | Drop connect rate |
  | ---------------- | ---------------- | ------- | ----------------- |
  | EfficientNetV1B0 | 224              | 0.2     | 0.2               |
  | EfficientNetV1B1 | 240              | 0.2     | 0.2               |
  | EfficientNetV1B2 | 260              | 0.3     | 0.2               |
  | EfficientNetV1B3 | 300              | 0.3     | 0.2               |
  | EfficientNetV1B4 | 380              | 0.4     | 0.2               |
  | EfficientNetV1B5 | 456              | 0.4     | 0.2               |
  | EfficientNetV1B6 | 528              | 0.5     | 0.2               |
  | EfficientNetV1B7 | 600              | 0.5     | 0.2               |
  | EfficientNetV1L2 | 800              | 0.5     | 0.2               |
"""

__v1_default_doc__ = __v1_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained=[None, "imagenet", "noisy_student"]) + """
Training configures: `Eval size` is used as the default model `input_shape`.
  | Model            | Input resolution | Dropout | Drop connect rate |
  | ---------------- | ---------------- | ------- | ----------------- |
  {train_config}
"""

EfficientNetV1B0.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B0 | 224              | 0.2     | 0.2               |")
EfficientNetV1B1.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B1 | 240              | 0.2     | 0.2               |")
EfficientNetV1B2.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B2 | 260              | 0.3     | 0.2               |")
EfficientNetV1B3.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B3 | 300              | 0.3     | 0.2               |")
EfficientNetV1B4.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B4 | 380              | 0.4     | 0.2               |")
EfficientNetV1B5.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B5 | 456              | 0.4     | 0.2               |")
EfficientNetV1B6.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B6 | 528              | 0.5     | 0.2               |")
EfficientNetV1B7.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1B7 | 600              | 0.5     | 0.2               |")
EfficientNetV1L2.__doc__ = __v1_default_doc__.format(train_config="| EfficientNetV1L2 | 800              | 0.5     | 0.2               |")

__v1_lite_default_doc__ = __v1_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained=[None]) + """
Training configures: `Eval size` is used as the default model `input_shape`.
  | Model               | Input resolution | Dropout | Drop connect rate |
  | ------------------- | ---------------- | ------- | ----------------- |
  {train_config}
"""
EfficientNetV1Lite0.__doc__ = __v1_lite_default_doc__.format(train_config="| EfficientNetV1Lite0 | 320              | 0.2     | 0.2               |")
EfficientNetV1Lite1.__doc__ = __v1_lite_default_doc__.format(train_config="| EfficientNetV1Lite1 | 384              | 0.2     | 0.2               |")
EfficientNetV1Lite2.__doc__ = __v1_lite_default_doc__.format(train_config="| EfficientNetV1Lite2 | 448              | 0.3     | 0.2               |")
EfficientNetV1Lite3.__doc__ = __v1_lite_default_doc__.format(train_config="| EfficientNetV1Lite3 | 512              | 0.3     | 0.2               |")
EfficientNetV1Lite4.__doc__ = __v1_lite_default_doc__.format(train_config="| EfficientNetV1Lite4 | 640              | 0.4     | 0.2               |")

__edge_tpu_head_doc__ = """
Keras implementation of [Github tensorflow/tpu/efficientnet/edgetpu](https://github.com/tensorflow/tpu/tree/master/models/official/efficientnet/edgetpu).
"""

EfficientNetEdgeTPU.__doc__ = __edge_tpu_head_doc__ + """
Args:
  model_name: string, model name.
""" + __tail_doc__.format(pretrained=[None, "imagenet"]) + """
Model architectures:
  | Edge TPU Model              | Params | FLOPs   | Input | Top1 Acc |
  | --------------------------- | ------ | ------- | ----- | -------- |
  | EfficientNetEdgeTPUSmall    | 5.49M  | 1.79G   | 224   | 78.07    |
  | EfficientNetEdgeTPUMedium   | 6.90M  | 3.01G   | 240   | 79.25    |
  | EfficientNetEdgeTPULarge    | 10.59M | 7.94G   | 300   | 81.32    |

Training configures:
  | Model                     | Input resolution | Dropout | Drop connect rate |
  | ------------------------- | ---------------- | ------- | ----------------- |
  | EfficientNetEdgeTPUSmall  | 224              | 0.2     | 0.2               |
  | EfficientNetEdgeTPUMedium | 240              | 0.2     | 0.2               |
  | EfficientNetEdgeTPULarge  | 300              | 0.3     | 0.2               |
"""

__edge_tpu_default_doc__ = __edge_tpu_head_doc__ + """
Args:
""" + __tail_doc__.format(pretrained=[None, "imagenet"]) + """
Training configures: `Eval size` is used as the default model `input_shape`.
  | Model                     | Input resolution | Dropout | Drop connect rate |
  | ------------------------- | ---------------- | ------- | ----------------- |
  {train_config}
"""

EfficientNetEdgeTPUSmall.__doc__ = __edge_tpu_default_doc__.format(train_config="| EfficientNetEdgeTPUSmall  | 224              | 0.2     | 0.2               |")
EfficientNetEdgeTPUMedium.__doc__ = __edge_tpu_default_doc__.format(train_config="| EfficientNetEdgeTPUMedium | 240              | 0.2     | 0.2               |")
EfficientNetEdgeTPULarge.__doc__ = __edge_tpu_default_doc__.format(train_config="| EfficientNetEdgeTPULarge  | 300              | 0.2     | 0.3               |")



"""
Wrappering a kecam PyTorch YOLOV8 model for training using ultralytics package
"""

import math
import torch
from torch import nn


def dist2bbox(distance, anchor_points, xywh=True, dim=-1):
    """Transform distance(ltrb) to box(xywh or xyxy)."""
    lt, rb = distance.chunk(2, dim)
    x1y1 = anchor_points - lt
    x2y2 = anchor_points + rb
    if xywh:
        c_xy = (x1y1 + x2y2) / 2
        wh = x2y2 - x1y1
        return torch.cat((c_xy, wh), dim)  # xywh bbox
    return torch.cat((x1y1, x2y2), dim)  # xyxy bbox


class Detect(nn.Module):
    def __init__(self, model, train_input_shape=640, reg_max=16, num_pyramid_levels=3, export=False):  # detection layer
        super().__init__()
        self.model, self.reg_max, self.num_pyramid_levels, self.export = model, reg_max, num_pyramid_levels, export
        self.device = next(self.model.parameters()).device

        train_input_shape = model.input_shape[2:] if hasattr(model, "input_shape") and model.input_shape[2] is not None else train_input_shape
        train_input_shape = train_input_shape[-2:] if isinstance(train_input_shape, (list, tuple)) else (train_input_shape, train_input_shape)
        self.input_shape, self.input_height, self.input_width = [None, 3, *train_input_shape], train_input_shape[0], train_input_shape[1]
        self.feature_sizes, self.feature_lens = self.get_feature_sizes(train_input_shape)

        self.output_shape = model.output_shape
        self.num_classes = self.output_shape[-1] - self.reg_max * 4
        self.names = {ii: str(ii) for ii in range(self.num_classes)}

        # self.dfl = DFL(self.reg_max) if self.reg_max > 1 else nn.Identity()
        self.dfl = torch.arange(self.reg_max, device=self.device, dtype=torch.float32).view(1, 1, self.reg_max, 1)
        self.pre_val_input_shape, self.pre_val_feature_sizes, self.pre_val_feature_lens = None, None, None

    def get_feature_sizes(self, input_shape):
        feature_sizes = [(math.ceil(input_shape[-2] / (2**ii)), math.ceil(input_shape[-1] / (2**ii))) for ii in range(3, 3 + self.num_pyramid_levels)]
        feature_lens = [ii[0] * ii[1] for ii in feature_sizes]
        return feature_sizes, feature_lens

    def make_anchors(self, feature_sizes, grid_cell_offset=0.5):
        """Generate anchors from features."""
        anchor_points, stride_tensor = [], []
        strides = [2**ii for ii in range(3, 3 + self.num_pyramid_levels)]
        for (hh, ww), stride in zip(feature_sizes, strides):
            sx = torch.arange(end=ww, device=self.device, dtype=torch.float32) + grid_cell_offset  # shift x
            sy = torch.arange(end=hh, device=self.device, dtype=torch.float32) + grid_cell_offset  # shift y
            sy, sx = torch.meshgrid(sy, sx, indexing="ij")
            anchor_points.append(torch.stack((sx, sy), -1).view(-1, 2))
            stride_tensor.append(torch.full((hh * ww, 1), stride, dtype=torch.float32, device=self.device))
        return torch.cat(anchor_points).transpose(0, 1), torch.cat(stride_tensor).transpose(0, 1)

    def forward(self, inputs):
        out = self.model(inputs)
        out = out.permute([0, 2, 1])

        if self.training:
            feature_sizes, feature_lens = self.feature_sizes, self.feature_lens
        elif inputs.shape[1:] == self.pre_val_input_shape:
            feature_sizes, feature_lens = self.pre_val_feature_sizes, self.pre_val_feature_lens
        else:
            feature_sizes, feature_lens = self.get_feature_sizes(inputs.shape[2:])
            self.anchors, self.strides = self.make_anchors(feature_sizes)
            self.pre_val_input_shape, self.pre_val_feature_sizes, self.pre_val_feature_lens = inputs.shape[1:], feature_sizes, feature_lens

        train_out = torch.split(out, feature_lens, dim=-1)
        train_out = [ii.view([-1, ii.shape[1], hh, ww]) for ii, (hh, ww) in zip(train_out, feature_sizes)]

        if self.training:
            return train_out
        else:
            box, cls = torch.split(out, (self.reg_max * 4, self.num_classes), dim=1)
            # box = box.reshape([-1, 4, self.reg_max, box.shape[-1]])[:, [1, 0, 3, 2]].reshape([-1, 64, box.shape[-1]])
            box = (box.view(-1, 4, self.reg_max, box.shape[-1]).softmax(2) * self.dfl).sum(2).view(-1, 4, box.shape[-1])
            box = dist2bbox(box, self.anchors.unsqueeze(0), xywh=True, dim=1) * self.strides
            val_out = torch.cat((box, cls.sigmoid()), 1)
            return val_out if self.export else (val_out, train_out)


if __name__ == "__main__":
    os.environ["KECAM_BACKEND"] = "torch"
    sys.path.append("../ultralytics/")
    import torch
    from keras_cv_attention_models.test_images import dog_cat
    from skimage.transform import resize

    tt = torch.load("yolov8n.pt")["model"]
    _ = tt.eval()
    _ = tt.float()

    imm = resize(dog_cat(), [640, 640])
    preds_torch, torch_out = tt(torch.from_numpy(imm[None]).permute([0, 3, 1, 2]).float())

    import torch
    from keras_cv_attention_models.yolov8 import torch_wrapper, yolov8
    from keras_cv_attention_models.test_images import dog_cat
    from skimage.transform import resize

    mm = yolov8.YOLOV8_N(classifier_activation=None, input_shape=(640, 640, 3))
    tt = torch_wrapper.Detect(mm)
    _ = tt.eval()
    imm = resize(dog_cat(), [640, 640])
    preds_torch, torch_out = tt(torch.from_numpy(imm[None]).permute([0, 3, 1, 2]).float())



class Optimizer:
    def __init__(
        self,
        name,
        weight_decay=0,
        clipnorm=None,
        clipvalue=None,
        global_clipnorm=None,
        use_ema=False,
        ema_momentum=0.99,
        ema_overwrite_frequency=None,
    ):
        pass



import os
import numpy as np
import tensorrt as trt
import pycuda.driver as cuda

try:
    # Use autoprimaryctx if available (pycuda >= 2021.1) to prevent issues with other modules that rely on the primary device context.
    import pycuda.autoprimaryctx
except ModuleNotFoundError:
    import pycuda.autoinit
TRT_LOGGER = trt.Logger(trt.Logger.INFO)


def init_mean_std_by_rescale_mode(rescale_mode):
    if isinstance(rescale_mode, (list, tuple)):  # Specific mean and std
        mean, std = rescale_mode
    elif rescale_mode == "torch":
        mean = np.array([0.485, 0.456, 0.406]).astype("float32") * 255.0
        std = np.array([0.229, 0.224, 0.225]).astype("float32") * 255.0
    elif rescale_mode == "tf":  # [0, 255] -> [-1, 1]
        mean, std = 127.5, 127.5
        # mean, std = 127.5, 128.0
    elif rescale_mode == "tf128":  # [0, 255] -> [-1, 1]
        mean, std = 128.0, 128.0
    elif rescale_mode == "raw01":
        mean, std = 0, 255.0  # [0, 255] -> [0, 1]
    else:
        mean, std = 0, 1  # raw inputs [0, 255]
    return mean, std


class ImageCalibrator(trt.IInt8EntropyCalibrator2):
    def __init__(self, data, cache_file, rescale_mode="torch"):
        super().__init__()
        self.data, self.cache_file, self.rescale_mode = data, cache_file, rescale_mode
        self.built = False

    def build(self, target_shape, batch_size, data_format="channels_last"):
        # Every time get_batch is called, the next batch of size batch_size will be copied to the device and returned.
        mean, std = init_mean_std_by_rescale_mode(self.rescale_mode)
        target_shape = target_shape if isinstance(target_shape, (list, tuple)) else [target_shape, target_shape]
        data = self.data
        if isinstance(data, str):
            from PIL import Image

            print(">>>> Read input data from path:", data)
            target_shape = list(target_shape)[::-1]  # [width, height] for Image.resize
            data = [np.array(Image.open(os.path.join(data, ii)).resize(target_shape)) for ii in os.listdir(data)]
        else:
            resize_shape = list(target_shape)[::-1]  # [width, height] for Image.resize
            target_shape = tuple(target_shape)
            data = [ii if tuple(ii.shape[:2]) == target_shape else np.array(Image.fromarray(ii).resize(resize_shape)) for ii in data]
        data = (np.array(data) - mean) / std
        if data_format != "channels_last":
            data = data.transpose([0, 3, 1, 2])
        self.built_data, self.data_format = data, data_format

        self.batch_size, self.target_shape = batch_size, target_shape
        self.current_index = 0
        self.device_input = cuda.mem_alloc(data[0].nbytes * self.batch_size)  # Allocate enough memory for a whole batch.
        self.built = True
        print(">>>> Built data info: data.shape={}, data.min={:.4f}, data.max={:.4f}".format(data.shape, data.min(), data.max()))

    def get_batch_size(self):
        return self.batch_size

    # TensorRT passes along the names of the engine bindings to the get_batch function.
    # You don't necessarily have to use them, but they can be useful to understand the order of
    # the inputs. The bindings list is expected to have the same ordering as 'names'.
    def get_batch(self, names):
        if self.current_index + self.batch_size > self.built_data.shape[0]:
            return None

        current_batch = int(self.current_index / self.batch_size)
        if current_batch % 10 == 0:
            print(">>>> Calibrating batch {:}, containing {:} images".format(current_batch, self.batch_size))

        batch = self.built_data[self.current_index : self.current_index + self.batch_size].ravel()
        cuda.memcpy_htod(self.device_input, batch)
        self.current_index += self.batch_size
        return [self.device_input]

    def read_calibration_cache(self):
        # If there is a cache, use it instead of calibrating again. Otherwise, implicitly return None.
        if os.path.exists(self.cache_file):
            with open(self.cache_file, "rb") as f:
                return f.read()

    def write_calibration_cache(self, cache):
        with open(self.cache_file, "wb") as f:
            f.write(cache)


def build_onnx_engine_one_input(model_file, engine_path=None, int8_calibrator=None, batch_size=-1, data_format="auto", max_workspace_size=-1):
    # model_file, engine_path, int8_calibrator, batch_size, data_format, max_workspace_size = "aaa.onnx", None, None, -1, "auto", -1
    EXPLICIT_BATCH = 1 << (int)(trt.NetworkDefinitionCreationFlag.EXPLICIT_BATCH)
    trt.init_libnvinfer_plugins(TRT_LOGGER, namespace="")

    builder = trt.Builder(TRT_LOGGER)
    config = builder.create_builder_config()
    max_workspace_size = max_workspace_size if max_workspace_size > 0 else 8
    # config.max_workspace_size = max_workspace_size * (2 ** 30)  # 8 GB
    config.set_memory_pool_limit(trt.MemoryPoolType.WORKSPACE, pool_size=max_workspace_size * (2**30))

    network = builder.create_network(EXPLICIT_BATCH)
    parser = trt.OnnxParser(network, TRT_LOGGER)
    with open(model_file, "rb") as model:
        parser.parse(model.read())
    assert parser.num_errors == 0

    inputs = network.get_input(0)
    input_shape = inputs.shape
    if batch_size > 0:
        input_shape = [batch_size] + list(input_shape[1:])
        network.get_input(0).shape = input_shape
    else:
        batch_size = inputs.shape[0]
    # builder.max_batch_size = batch_size
    print(">>>> Input name: {}, shape: {}, dtype: {}".format(inputs.name, input_shape, inputs.dtype))

    outputs = [network.get_output(ii) for ii in range(network.num_outputs)]
    for output in outputs:
        print(">>>> Output name: {}, shape: {}, dtype: {}".format(output.name, output.shape, output.dtype))

    if int8_calibrator is None:
        config.set_flag(trt.BuilderFlag.FP16)
    else:
        config.set_flag(trt.BuilderFlag.INT8)
        if data_format == "auto":
            data_format = "channels_last" if input_shape[-1] < input_shape[1] else "channels_first"
        data_input_shape = input_shape[1:-1] if data_format == "channels_last" else input_shape[2:]
        print("data_format = {}, data_input_shape = {}".format(data_format, data_input_shape))
        int8_calibrator.build(data_input_shape, batch_size, data_format=data_format)
        config.int8_calibrator = int8_calibrator

    # Dynamic batch_size
    # profile = builder.create_optimization_profile()
    # input_shape = network.get_input(0).shape[1:]
    # profile.set_shape(network.get_input(0).name, (1, *input_shape), (1, *input_shape), (builder.max_batch_size, *input_shape))
    # config.add_optimization_profile(profile)

    # engine = builder.build_engine(network, config)
    engine = builder.build_serialized_network(network, config)
    if engine_path is None:
        engine_path = "{}_{}.trt".format(os.path.splitext(model_file)[0], "float16" if int8_calibrator is None else "int8")
    with open(engine_path, "wb") as ff:
        print(">>>> Serializing engine to file: {:}".format(engine_path))
        ff.write(engine)
    return engine


class EngineInferenceOneInOneOut:
    """
    >>> !pip install tensorrt pycuda
    >>> import torch
    >>> from keras_cv_attention_models.imagenet import tensorrt_engine
    >>> aa = tensorrt_engine.ImageCalibrator('calibration_imagenet/', 'foo.cache')
    >>> ee = tensorrt_engine.build_onnx_engine_one_input('aaa.onnx', int8_calibrator=aa)
    >>> cc = tensorrt_engine.EngineInferenceOneInOneOut(ee, max_batch_size=4)
    >>> print(cc(np.ones([1, 3, 224, 224])).shape)
    >>> # (1, 1000)
    >>> print(cc(np.ones([4, 3, 224, 224])).shape)
    >>> # (4, 1000)
    """

    def __init__(self, engine, max_batch_size=1):
        if isinstance(engine, str):
            with open(engine_path, "rb") as ff:
                engine = ff.read()
        with trt.Runtime(TRT_LOGGER) as runtime:
            engine = runtime.deserialize_cuda_engine(engine)
        assert runtime

        input_binding, output_binding = None, None
        for binding in engine:
            if engine.get_tensor_mode(binding) == trt.TensorIOMode.INPUT:
                input_binding = binding
            else:
                output_binding = binding
        assert input_binding
        assert output_binding

        self.input_shape = [max_batch_size, *engine.get_tensor_shape(input_binding)[1:]]
        self.input_dtype = trt.nptype(engine.get_tensor_dtype(input_binding))
        size = trt.volume(engine.get_tensor_shape(input_binding)) * max_batch_size
        self.host_input = cuda.pagelocked_empty(shape=[size], dtype=self.input_dtype)
        self.cuda_input = cuda.mem_alloc(self.host_input.nbytes)
        self.input_binding, self.output_binding = input_binding, output_binding

        self.output_shape = [max_batch_size, *engine.get_tensor_shape(output_binding)[1:]]
        self.output_dtype = trt.nptype(engine.get_tensor_dtype(output_binding))
        size = trt.volume(engine.get_tensor_shape(output_binding)) * max_batch_size
        self.host_output = cuda.pagelocked_empty(shape=[size], dtype=self.output_dtype)
        self.cuda_output = cuda.mem_alloc(self.host_output.nbytes)

        self.output_dim = self.output_shape[1:]
        self.output_ravel_dim = self.host_output.shape[0] // max_batch_size
        self.allocations = [int(self.cuda_input), int(self.cuda_output)]
        self.max_batch_size = max_batch_size

        self.engine = engine
        self.stream = cuda.Stream()
        self.context = engine.create_execution_context()

    def __call__(self, imgs):
        batch_size = imgs.shape[0]
        if batch_size > self.max_batch_size:
            print(f"Warning: provided input with batch_size={batch_size} exceeds max_batch_size={self.max_batch_size}")
            batch_size = self.max_batch_size
            imgs = imgs[: self.max_batch_size]

        inputs = imgs.ravel()
        # self.context.set_binding_shape(0, imgs.shape)
        np.copyto(self.host_input[: inputs.shape[0]], inputs)
        cuda.memcpy_htod_async(self.cuda_input, self.host_input[: inputs.shape[0]], self.stream)
        # Run inference asynchronously, same function in cpp is `IExecutionContext::enqueueV2`
        self.context.execute_async_v2(bindings=self.allocations, stream_handle=self.stream.handle)
        # Transfer predictions back from the GPU.
        cuda.memcpy_dtoh_async(self.host_output[: batch_size * self.output_ravel_dim], self.cuda_output, self.stream)
        # Synchronize the stream
        self.stream.synchronize()
        return self.host_output[: batch_size * self.output_ravel_dim].reshape([batch_size, *self.output_dim]).copy()


if __name__ == "__main__":
    # !pip install kecam torch torchvision onnx pycuda tensorrt
    # !git clone https://github.com/NVIDIA/TensorRT.git
    # cd TensorRT/samples/trtexec

    import torch
    import torchvision
    from kecam.imagenet.tensorrt_engine import ImageCalibrator, build_onnx_engine_one_input, EngineInferenceOneInOneOut

    mm = torchvision.models.resnet50(pretrained=True)
    torch.onnx.export(mm, torch.ones([1, 3, 224, 224]), "aaa.onnx")
    aa = ImageCalibrator("calibration_imagenet/", "foo.cache")
    ee = build_onnx_engine_one_input("aaa.onnx", int8_calibrator=aa)
    cc = EngineInferenceOneInOneOut(ee)
    print(cc(np.ones([1, 3, 224, 224])).shape)



from keras_cv_attention_models.nat.nat import (
    NAT,
    NAT_Mini,
    NAT_Tiny,
    NAT_Small,
    NAT_Base,
    MultiHeadRelativePositionalKernelBias,
    neighborhood_attention,
)
from keras_cv_attention_models.nat.dinat import (
    DiNAT_Mini,
    DiNAT_Tiny,
    DiNAT_Small,
    DiNAT_Base,
    DiNAT_Large,
    DiNAT_Large_K11,
)

__head_doc__ = """
Keras implementation of [NAT](https://github.com/SHI-Labs/Neighborhood-Attention-Transformer).
Paper [PDF 2204.07143 Neighborhood Attention Transformer](https://arxiv.org/pdf/2204.07143.pdf).
Paper [PDF 2209.15001 Dilated Neighborhood Attention Transformer](https://arxiv.org/pdf/2209.15001.pdf).
"""

__tail_doc__ = """  stem_width: output dimension for stem block. Default -1 means using `out_channels[0]`
  attn_kernel_size: kernel_size for `neighborhood_attention` block, defualt 7.
  use_every_other_dilations: True for DiNAT, False for others.
      Using `dilation_rate=nn.shape[1] // attn_kernel_size` in every other attention blocks.
  mlp_ratio: channel expansion ratio for mlp hidden layers, default 3 for NAT_Mini and NAT_Tiny, 2 for NAT_Small and NAT_Base.
  layer_scale: layer scale init value. `-1` means not applying, any value `>=0` will add a scale value for each block output.
      [Going deeper with Image Transformers](https://arxiv.org/pdf/2103.17239.pdf).
      Default -1 for NAT_Mini and NAT_Tiny, 1e-5 for NAT_Small and NAT_Base
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `gelu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  dropout: dropout rate if top layers is included.
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
  pretrained: None or one of ["imagenet21k-ft1k", "imagenet21k"] for `DiNAT_Large`, or "imagenet21k-ft1k" for `DiNAT_Large_K11`,
      or "imagenet" for others.

Returns:
    A `keras.Model` instance.
"""

NAT.__doc__ = __head_doc__ + """
Args:
  num_blocks: number of blocks in each stack.
  out_channels: output channels for each stack.
  num_heads: num heads for each stack.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model     | Params | FLOPs  | Input | Top1 Acc |
  | --------- | ------ | ------ | ----- | -------- |
  | NAT_Mini  | 20.0M  | 2.73G  | 224   | 81.8     |
  | NAT_Tiny  | 27.9M  | 4.34G  | 224   | 83.2     |
  | NAT_Small | 50.7M  | 7.84G  | 224   | 83.7     |
  | NAT_Base  | 89.8M  | 13.76G | 224   | 84.3     |

  | Model                     | Params | FLOPs  | Input | Top1 Acc |
  | ------------------------- | ------ | ------ | ----- | -------- |
  | DiNAT_Mini                | 20.0M  | 2.73G  | 224   | 81.8     |
  | DiNAT_Tiny                | 27.9M  | 4.34G  | 224   | 82.7     |
  | DiNAT_Small               | 50.7M  | 7.84G  | 224   | 83.8     |
  | DiNAT_Base                | 89.8M  | 13.76G | 224   | 84.4     |
  | DiNAT_Large, 22k          | 200.9M | 30.58G | 224   | 86.6     |
  | - 21k num_classes=21841   | 200.9M | 30.58G | 224   |          |
  | - 22k, 384                | 200.9M | 89.86G | 384   | 87.4     |
  | DiNAT_Large_K11, 22k, 384 | 201.1M | 92.57G | 384   | 87.5     |
"""

NAT_Mini.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

NAT_Tiny.__doc__ = NAT_Mini.__doc__
NAT_Small.__doc__ = NAT_Mini.__doc__
NAT_Base.__doc__ = NAT_Mini.__doc__

DiNAT_Mini.__doc__ = NAT_Mini.__doc__
DiNAT_Tiny.__doc__ = NAT_Mini.__doc__
DiNAT_Small.__doc__ = NAT_Mini.__doc__
DiNAT_Base.__doc__ = NAT_Mini.__doc__
DiNAT_Large.__doc__ = NAT_Mini.__doc__
DiNAT_Large_K11.__doc__ = NAT_Mini.__doc__

MultiHeadRelativePositionalKernelBias.__doc__ = __head_doc__ + """
Multi Head Relative Positional Kernel Bias layer. Weights depends on `num_heads` and `kernel_size` not on `input_shape`.

input (is_heads_first=False): `[batch, height * width, num_heads, ..., size * size]`
input (is_heads_first=True): `[batch, num_heads, height * width, ..., size * size]`
positional_bias: `[num_heads, (2 * size - 1) * (2 * size - 1)]`
output: `input + gathered positional_bias`.
condition: height >= size, width >= size

Args:
  input_height: specify `height` for `input` if not square, default -1 assums `input_height == input_width`.
  is_heads_first: boolean value if input is `[batch, num_heads, height * width]` or `[batch, height * width, num_heads]`.
  dilation_rate: dilation_rate used for last dimension `size * size`.

Examples:

# Basic
>>> from keras_cv_attention_models import attention_layers
>>> aa = attention_layers.MultiHeadRelativePositionalKernelBias()
>>> print(f"{aa(tf.ones([1, 29 * 29, 8, 5 * 5])).shape = }")
# aa(tf.ones([1, 29 * 29, 8, 5 * 5])).shape = TensorShape([1, 841, 8, 25])
>>> print(f'{aa.pos_bias.shape = }, {aa.bias_coords.shape = }')
# aa.pos_bias.shape = TensorShape([8, 81]), aa.bias_coords.shape = TensorShape([841, 25])

# Specific input_height and a longer input shape
>>> aa = attention_layers.MultiHeadRelativePositionalKernelBias(input_height=7)
>>> print(f"{aa(tf.ones([1, 7 * 13, 8, 23, 46, 4, 3 * 3])).shape = }")
# aa(tf.ones([1, 7 * 13, 8, 23, 46, 4, 3 * 3])).shape = TensorShape([1, 91, 8, 23, 46, 4, 9])
>>> print(f'{aa.pos_bias.shape = }, {aa.bias_coords.shape = }')
# aa.pos_bias.shape = TensorShape([8, 25]), aa.bias_coords.shape = TensorShape([91, 1, 1, 1, 9])

# Specific is_heads_first=True
>>> aa = attention_layers.MultiHeadRelativePositionalKernelBias(input_height=19, is_heads_first=True)
>>> print(f"{aa(tf.ones([1, 8, 19 * 29, 5 * 5])).shape = }")
# aa(tf.ones([1, 8, 19 * 29, 5 * 5])).shape = TensorShape([1, 8, 551, 25])
>>> print(f'{aa.pos_bias.shape = }, {aa.bias_coords.shape = }')
# aa.pos_bias.shape = TensorShape([8, 81]), aa.bias_coords.shape = TensorShape([551, 25])
"""

neighborhood_attention.__doc__ = __head_doc__ + """
Neighborhood Attention block, not a layer.
Extract patches with a `kernel_size` from `key_value` as an enlarged attention area. Balancing global and local attention.
Also adds `MultiHeadRelativePositionalKernelBias` with `attention_scores`.

Args:
  inputs: input tensor.
  kernel_size: extracting patch kernel_size for key and value.
  num_heads: Number of attention heads.
  key_dim: Size of each attention head for query and key. Default `0` for `key_dim = inputs.shape[-1] // num_heads`.
  out_weight: Boolean, whether use an ouput dense.
  qkv_bias: Boolean, whether the qkv dense layer use bias vectors/matrices.
  out_bias: Boolean, whether the ouput dense layer use bias vectors/matrices.
  dilation_rate: dilation_rate for extracting `key` and `value` features.
  attn_dropout: Dropout probability for attention scores.
  output_dropout: Dropout probability for attention output.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([14, 16, 256])
>>> nn = attention_layers.neighborhood_attention(inputs, num_heads=4)
>>> print(f"{nn.shape = }")
# nn.shape = TensorShape([None, 14, 16, 512])

>>> mm = keras.models.Model(inputs, nn)
>>> mm.summary()
>>> print({ii.name: ii.shape for ii in mm.weights})
# {'dense/kernel:0': TensorShape([256, 768]),
#  'dense/bias:0': TensorShape([768]),
#  'multi_head_relative_positional_kernel_bias_11/positional_embedding:0': TensorShape([4, 169]),
#  'dense_1/kernel:0': TensorShape([256, 256]),
#  'dense_1/bias:0': TensorShape([256])}
"""



import os
import torch
import numpy as np
from PIL import Image
from torch.utils.data import Dataset, DataLoader
from torchvision.transforms import Normalize, Compose, RandomResizedCrop, Resize, InterpolationMode, ToTensor


def read_from_tsv(data_path):
    import csv

    delimiter = "\t" if data_path.endswith(".tsv") else ","
    train_images, train_captions, test_images, test_captions, base_path, is_train = [], [], [], [], "", True
    with open(data_path) as ff:
        for ii in csv.reader(ff, delimiter=delimiter):
            if ii[0] == "base_path":  # special keys for info
                base_path = os.path.expanduser(ii[1])
            elif ii[0] == "TEST":  # Use this as indicator for start of test set
                is_train = False
            elif is_train:
                train_images.append(ii[0])
                train_captions.append(ii[1])
            else:
                test_images.append(ii[0])
                test_captions.append(ii[1])
    if len(base_path) > 0:
        train_images = [os.path.join(base_path, ii) for ii in train_images]
        test_images = [os.path.join(base_path, ii) for ii in test_images]
    return train_images, train_captions, test_images, test_captions


class CaptionDataset(Dataset):
    def __init__(self, images, captions, tokenizer, is_train=True, image_size=224):
        self.images, self.captions, self.tokenizer = images, captions, tokenizer
        self.context_length = self.tokenizer.context_length

        # self.mean, self.std = (0.48145466, 0.4578275, 0.40821073), (0.26862954, 0.26130258, 0.27577711)  # value from openai/CLIP
        self.mean, self.std = (0.485, 0.456, 0.406), (0.229, 0.224, 0.225)
        interpolation = InterpolationMode.BICUBIC
        image_size = image_size if isinstance(image_size, (list, tuple)) else (image_size, image_size)
        self.transforms = Compose(
            [
                RandomResizedCrop(image_size, scale=(0.9, 1.0), interpolation=interpolation) if is_train else Resize(image_size, interpolation=interpolation),
                lambda image: image.convert("RGB"),
                ToTensor(),
                Normalize(mean=self.mean, std=self.std),
            ]
        )

    def __len__(self):
        return len(self.images)

    def __getitem__(self, idx):
        images = self.transforms(Image.open(str(self.images[idx])))
        texts = torch.from_numpy(self.tokenizer(str(self.captions[idx])))
        return images, texts


def collate_wrapper(batch):
    images, texts = list(zip(*batch))
    return (torch.stack(images), torch.stack(texts)), torch.arange(len(batch))


def init_dataset(data_path, caption_tokenizer, batch_size=64, image_size=224, num_workers=8):
    train_images, train_captions, test_images, test_captions = read_from_tsv(data_path)

    train_dataset = CaptionDataset(train_images, train_captions, tokenizer=caption_tokenizer, is_train=True, image_size=image_size)
    train_dataloader = DataLoader(
        train_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, collate_fn=collate_wrapper, pin_memory=True, sampler=None, drop_last=True
    )

    test_dataset = CaptionDataset(test_images, test_captions, tokenizer=caption_tokenizer, is_train=False, image_size=image_size)
    test_dataloader = DataLoader(
        test_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, collate_fn=collate_wrapper, pin_memory=True, sampler=None, drop_last=True
    )

    return train_dataloader, test_dataloader



import os
import time
import keras_cv_attention_models
from keras_cv_attention_models import backend, model_surgery
from keras_cv_attention_models.backend import layers, models
from keras_cv_attention_models.imagenet import callbacks

GLOBAL_STRATEGY = None


def init_global_strategy(enable_float16=True, seed=0, TPU=False):
    import tensorflow as tf

    global GLOBAL_STRATEGY
    if GLOBAL_STRATEGY is not None:
        return GLOBAL_STRATEGY

    gpus = tf.config.experimental.get_visible_devices("GPU")
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)

    if TPU:
        resolver = tf.distribute.cluster_resolver.TPUClusterResolver(tpu="")
        tf.config.experimental_connect_to_cluster(resolver)
        # This is the TPU initialization code that has to be at the beginning.
        tf.tpu.experimental.initialize_tpu_system(resolver)
        print("[TPU] All devices: ", tf.config.list_logical_devices("TPU"))
        strategy = tf.distribute.TPUStrategy(resolver)
    elif len(gpus) > 1:
        strategy = tf.distribute.MirroredStrategy()
    else:
        strategy = tf.distribute.OneDeviceStrategy(device="/gpu:0")
    GLOBAL_STRATEGY = strategy

    if enable_float16:
        policy = "mixed_bfloat16" if TPU else "mixed_float16"
        tf.keras.mixed_precision.set_global_policy(policy)

    if seed is not None:
        print(">>>> Set random seed:", seed)
        tf.random.set_seed(seed)
    return strategy


def init_lr_scheduler(lr_base, lr_decay_steps, lr_min=1e-5, lr_decay_on_batch=False, lr_warmup=1e-4, warmup_steps=0, cooldown_steps=0, t_mul=2, m_mul=0.5):
    if isinstance(lr_decay_steps, list):
        constant_lr_sch = lambda epoch: callbacks.constant_scheduler(epoch, lr_base=lr_base, lr_decay_steps=lr_decay_steps, warmup_steps=warmup_steps)
        lr_scheduler = callbacks.LearningRateScheduler(constant_lr_sch)
        lr_total_epochs = lr_decay_steps[-1] + cooldown_steps  # 120 for lr_decay_steps=[30, 60, 90], warmup_steps=4, cooldown_steps=30
    elif lr_decay_on_batch:
        lr_scheduler = callbacks.CosineLrScheduler(
            lr_base, lr_decay_steps, m_mul=m_mul, t_mul=t_mul, lr_min=lr_min, lr_warmup=lr_warmup, warmup_steps=warmup_steps, cooldown_steps=cooldown_steps
        )
        lr_total_epochs = lr_decay_steps + cooldown_steps  # 105 for lr_decay_steps=100, warmup_steps=4, cooldown_steps=5
    else:
        lr_scheduler = callbacks.CosineLrSchedulerEpoch(
            lr_base, lr_decay_steps, m_mul=m_mul, t_mul=t_mul, lr_min=lr_min, lr_warmup=lr_warmup, warmup_steps=warmup_steps, cooldown_steps=cooldown_steps
        )
        lr_total_epochs = lr_decay_steps + cooldown_steps  # 105 for lr_decay_steps=100, warmup_steps=4, cooldown_steps=5
    return lr_scheduler, lr_total_epochs


def _init_tf_buildin_optimizer_(optimizer_class, lr_base, weight_decay, no_weight_decay, momentum=0.9, **kwargs):
    import inspect

    is_weight_decay_supported = "weight_decay" in inspect.signature(optimizer_class).parameters  # > TF1.11
    if is_weight_decay_supported:
        kwargs.update({"weight_decay": weight_decay})
    if "momentum" in inspect.signature(optimizer_class).parameters:  # SGD / RMSprop
        kwargs.update({"momentum": momentum})
    optimizer = optimizer_class(learning_rate=lr_base, **kwargs)

    if is_weight_decay_supported:
        optimizer.exclude_from_weight_decay(var_names=no_weight_decay)
    return optimizer


def init_optimizer(optimizer, lr_base, weight_decay, momentum=0.9):
    optimizer = optimizer.lower()
    buildin_optimizers = {
        # key: (class, kwargs)
        "sgd": (backend.optimizers.SGD, {}),
        "rmsprop": (backend.optimizers.RMSprop, {}),
        "adam": (backend.optimizers.Adam, {}),
        "custom": (backend.optimizers.AdamW, {"beta_1": 0.9, "beta_2": 0.98, "epsilon": 1e-6, "global_clipnorm": 10.0}),  # For clip
    }
    if hasattr(backend.optimizers, "AdamW"):
        buildin_optimizers.update({"adamw": (backend.optimizers.AdamW, {})})
    # norm_weights = ["bn/gamma", "bn/beta", "ln/gamma", "ln/beta", "/positional_embedding", "/bias"]  # ["bn/moving_mean", "bn/moving_variance"] not in weights
    no_weight_decay = ["/gamma", "/beta", "/bias", "/positional_embedding", "/no_weight_decay"]  # ["bn/moving_mean", "bn/moving_variance"] not in weights

    if optimizer in buildin_optimizers:
        optimizer_class, kwargs = buildin_optimizers[optimizer]
        optimizer = _init_tf_buildin_optimizer_(optimizer_class, lr_base, weight_decay, no_weight_decay, momentum, **kwargs)
    elif optimizer == "lamb":
        import tensorflow_addons as tfa

        optimizer = tfa.optimizers.LAMB(learning_rate=lr_base, weight_decay_rate=weight_decay, exclude_from_weight_decay=no_weight_decay, global_clipnorm=1.0)
    elif optimizer == "adamw":
        import tensorflow_addons as tfa

        optimizer = tfa.optimizers.AdamW(learning_rate=lr_base, weight_decay=lr_base * weight_decay, global_clipnorm=1.0)
        if hasattr(optimizer, "exclude_from_weight_decay"):
            setattr(optimizer, "exclude_from_weight_decay", no_weight_decay)
    elif optimizer == "sgdw":
        import tensorflow_addons as tfa

        optimizer = tfa.optimizers.SGDW(learning_rate=lr_base, momentum=momentum, weight_decay=lr_base * weight_decay)
        if hasattr(optimizer, "exclude_from_weight_decay"):
            setattr(optimizer, "exclude_from_weight_decay", no_weight_decay)
    else:
        optimizer = getattr(backend.optimizers, optimizer.capitalize())(learning_rate=lr_base)

    return optimizer


def is_decoupled_weight_decay(optimizer):
    # import tensorflow_addons as tfa

    optimizer = optimizer.inner_optimizer if hasattr(optimizer, "inner_optimizer") else optimizer
    # return isinstance(optimizer.__module__, tfa.optimizers.weight_decay_optimizers.DecoupledWeightDecayExtension)
    return optimizer.__module__ == "tensorflow_addons.optimizers.weight_decay_optimizers"


def init_loss(bce_threshold=1.0, label_smoothing=0, token_label_loss_weight=0, distill_loss_weight=0, distill_temperature=10, model_output_names=[]):
    from keras_cv_attention_models.imagenet import losses

    from_logits = True if distill_loss_weight > 0 else False  # classifier_activation is set to None for distill model, set from_logits=True for distill
    if bce_threshold >= 0 and bce_threshold < 1:
        cls_loss = losses.BinaryCrossEntropyTimm(target_threshold=bce_threshold, label_smoothing=label_smoothing, from_logits=from_logits)
        aux_loss = losses.BinaryCrossEntropyTimm(target_threshold=bce_threshold, label_smoothing=label_smoothing) if token_label_loss_weight > 0 else None
    else:
        cls_loss = backend.losses.CategoricalCrossentropy(label_smoothing=label_smoothing, from_logits=from_logits)
        aux_loss = backend.losses.CategoricalCrossentropy(label_smoothing=label_smoothing) if token_label_loss_weight > 0 else None

    if token_label_loss_weight > 0:
        loss = [cls_loss, aux_loss]
        loss_weights = [1, token_label_loss_weight]
        metrics = {model_output_names[0]: "acc", model_output_names[1]: None}
    elif distill_loss_weight > 0:
        distill_loss = losses.DistillKLDivergenceLoss(temperature=distill_temperature)
        loss = [cls_loss, distill_loss]
        loss_weights = [1, distill_loss_weight]
        metrics = {model_output_names[0]: "acc", model_output_names[1]: None}
    else:
        loss = cls_loss
        loss_weights, metrics = None, ["acc"]
    return loss, loss_weights, metrics


def init_model(model=None, input_shape=(224, 224, 3), num_classes=1000, pretrained=None, reload_compile=True, **kwargs):
    """model Could be:
    1. Saved h5 model path.
    2. Model name defined in this repo, format [sub_dir].[model_name] like regnet.RegNetZD8.
    3. timm model like timm.models.resmlp_12_224
    """
    if isinstance(model, models.Model):
        print(">>> Got a models.Model: {}, do nothing with it.".format(model.name))
        return model

    if model.startswith("timm."):  # model like: timm.models.resmlp_12_224
        import timm
        from keras_cv_attention_models.imagenet.eval_func import TorchModelInterf

        print(">>>> Timm model provided:", model)
        timm_model_name = ".".join(model.split(".")[2:])
        model = getattr(timm.models, timm_model_name)(pretrained=True, img_size=input_shape[:2], num_classes=num_classes)
        return TorchModelInterf(model)

    if model.endswith(".h5"):
        try:
            import tensorflow_addons as tfa
        except:
            pass

        print(">>>> Restore model from:", model)
        model = models.load_model(model, compile=reload_compile)
        return model

    if input_shape != -1:
        kwargs.update({"input_shape": input_shape})  # Use model default input_shape if not specified
    if num_classes != -1:
        kwargs.update({"num_classes": num_classes})  # Use model default input_shape if not specified
    if pretrained != "default":
        kwargs.update({"pretrained": pretrained})  # Use model default input_shape if not specified
    print(">>>> init_model kwargs:", kwargs)

    model_name = model.strip().split(".")
    if len(model_name) == 1:
        model = getattr(keras_cv_attention_models.models, model_name[0])(**kwargs)
    else:
        model_class = getattr(getattr(keras_cv_attention_models, model_name[0]), model_name[1])
        model = model_class(**kwargs)
    print(">>>> Built model name:", model.name)

    if model_name[0] == "aotnet" and pretrained is not None and pretrained.endswith(".h5"):
        # Currently aotnet not loading from pretrained...
        print(">>>> Load pretrained from:", pretrained)
        model.load_weights(pretrained, by_name=True, skip_mismatch=True)
    return model


def model_post_process(model, freeze_backbone=False, freeze_norm_layers=False, use_token_label=False):
    if freeze_backbone:
        pool_layer_id = model_surgery.get_global_avg_pool_layer_id(model)
        for id in range(pool_layer_id):
            model.layers[id].trainable = False

    if freeze_norm_layers:
        for ii in model.layers:
            if isinstance(ii, layers.BatchNormalization) or isinstance(ii, layers.LayerNormalization):
                ii.trainable = False

    if use_token_label and model.optimizer is None:  # model.optimizer is not None if restored from h5
        model = model_surgery.convert_to_token_label_model(model)
    return model


def init_distill_model(model, teacher_model):
    if hasattr(teacher_model, "layers") and hasattr(teacher_model.layers[-1], "activation"):
        teacher_model.layers[-1].activation = None  # Set output activation softmax to linear
    teacher_model.trainable = False

    model.layers[-1].activation = None  # Also set model output activation softmax to linear
    if model.optimizer is None:  # model.optimizer is not None if restored from h5
        model = models.Model(model.inputs[0], [model.output, model.output])
        model.output_names[1] = "distill"
    return model, teacher_model


def compile_model(model, optimizer, lr_base, weight_decay, loss, loss_weights=None, metrics=["acc"], momentum=0.9):
    if isinstance(optimizer, str):
        optimizer = optimizer.lower()
        # if optimizer == "sgd" and weight_decay > 0:
        #     Add L2 regularizer
        #     model = model_surgery.add_l2_regularizer_2_model(model, weight_decay=weight_decay, apply_to_batch_normal=False)
        optimizer = init_optimizer(optimizer, lr_base, weight_decay, momentum=momentum)
    print(">>>> Loss: {}, Optimizer: {}".format(loss.__class__.__name__, optimizer.__class__.__name__))
    model.compile(optimizer=optimizer, loss=loss, loss_weights=loss_weights, metrics=metrics)

    if not hasattr(model.optimizer, "_variables") and hasattr(model.optimizer, "_optimizer") and hasattr(model.optimizer._optimizer, "_variables"):
        # Bypassing TF 2.11 error AttributeError: 'LossScaleOptimizerV3' object has no attribute '_variables'
        setattr(model.optimizer, "_variables", model.optimizer._optimizer._variables)
    return model


# @tf.function(jit_compile=True)
def train(
    compiled_model, epochs, train_dataset, test_dataset=None, initial_epoch=0, lr_scheduler=None, basic_save_name=None, init_callbacks=[], logs="auto", **kwargs
):
    if compiled_model.compiled_loss is None:
        print(">>>> Error: Model NOT compiled.")
        return None

    steps_per_epoch = len(train_dataset)
    if hasattr(lr_scheduler, "steps_per_epoch") and lr_scheduler.steps_per_epoch == -1:
        lr_scheduler.build(steps_per_epoch)
    is_lr_on_batch = True if hasattr(lr_scheduler, "steps_per_epoch") and lr_scheduler.steps_per_epoch > 0 else False

    if basic_save_name is None:
        basic_save_name = "{}".format(compiled_model.name)
    # ckpt_path = os.path.join("checkpoints", basic_save_name + "epoch_{epoch:03d}_val_acc_{val_acc:.4f}.h5")
    # cur_callbacks = [keras.callbacks.ModelCheckpoint(ckpt_path, monitor="val_loss", save_best_only=True)]
    # cur_callbacks = [keras.callbacks.ModelCheckpoint(os.path.join("checkpoints", basic_save_name + ".h5"))]
    checkpoint_callback = callbacks.MyCheckpoint(basic_save_name, monitor="val_acc")
    cur_callbacks = [checkpoint_callback] + init_callbacks
    hist_file = os.path.join("checkpoints", basic_save_name + "_hist.json")
    if initial_epoch == 0 and os.path.exists(hist_file):
        # os.remove(hist_file)
        os.rename(hist_file, hist_file + ".bak")
    cur_callbacks.append(callbacks.MyHistory(initial_file=hist_file))
    cur_callbacks.append(backend.callbacks.TerminateOnNaN())
    if logs is not None:
        logs = "logs/" + basic_save_name + "_" + time.strftime("%Y%m%d-%H%M%S") if logs == "auto" else logs
        cur_callbacks.append(backend.callbacks.TensorBoard(log_dir=logs, histogram_freq=1))
        print(">>>> TensorBoard log path:", logs)

    if lr_scheduler is not None:
        cur_callbacks.append(lr_scheduler)

    if lr_scheduler is not None and is_decoupled_weight_decay(compiled_model.optimizer):
        print(">>>> Append weight decay callback...")
        lr_base, wd_base = compiled_model.optimizer.lr.numpy(), compiled_model.optimizer.weight_decay.numpy()
        wd_callback = callbacks.OptimizerWeightDecay(lr_base, wd_base, is_lr_on_batch=is_lr_on_batch)
        cur_callbacks.append(wd_callback)  # should be after lr_scheduler

    hist = compiled_model.fit(
        train_dataset,
        epochs=epochs,
        verbose=1,
        callbacks=cur_callbacks,
        initial_epoch=initial_epoch,
        steps_per_epoch=steps_per_epoch,
        validation_data=test_dataset,
        use_multiprocessing=True,
        workers=8,
        **kwargs,
    )

    if logs is not None:
        print(">>>> TensorBoard log path:", logs)
    return checkpoint_callback.latest_save, hist



import os
import tensorflow as tf
from tensorflow import keras
from keras_cv_attention_models.imagenet.data import init_mean_std_by_rescale_mode, tf_imread, random_crop_and_resize_image
from keras_cv_attention_models.coco import anchors_func
from keras_cv_attention_models.plot_func import draw_bboxes, show_image_with_bboxes


""" Bboxes augment """


def rerange_scale_offset_to_01(source_height, source_width, target_height, target_width, scale_hh, scale_ww, offset_hh=0, offset_ww=0):
    # Input: image size firstly rescale with (scale_hh, scale_ww), then crop as [offset_hh: offset_hh + target_height, offset_ww: offset_ww + target_width]
    # Output: coords or bbox value in (0, 1) -> * scale - offset
    source_height, source_width = tf.cast(source_height, "float32"), tf.cast(source_width, "float32")
    target_height, target_width = tf.cast(target_height, "float32"), tf.cast(target_width, "float32")
    scale_hh_01, scale_ww_01 = source_height * scale_hh / target_height, source_width * scale_ww / target_width
    crop_hh_01, crop_ww_01 = tf.cast(offset_hh, "float32") / target_height, tf.cast(offset_ww, "float32") / target_width
    return scale_hh_01, scale_ww_01, crop_hh_01, crop_ww_01


def inverse_affine_matrix_single_6(affine):
    scale_inverse = affine[0] * affine[4] - affine[3] * affine[1]
    affine_aa = [affine[4], -affine[1], affine[5] * affine[1] - affine[2] * affine[4]]
    affine_bb = [-affine[3], affine[0], affine[2] * affine[3] - affine[5] * affine[0]]
    return tf.convert_to_tensor(affine_aa + affine_bb, dtype=affine.dtype) / scale_inverse


def bboxes_apply_affine(bboxes, affine, input_shape, inverse=True):
    height, width = tf.cast(input_shape[0], bboxes.dtype), tf.cast(input_shape[1], bboxes.dtype)
    top, left, bottom, right = bboxes[:, 0] * height, bboxes[:, 1] * width, bboxes[:, 2] * height, bboxes[:, 3] * width
    affine = tf.squeeze(affine)[:6]
    if inverse:
        affine = inverse_affine_matrix_single_6(affine)
    affine = tf.reshape(affine, [2, 3])

    # 4 corners points: [left, top, right, bottom, left, bottom, right, top]
    corners = tf.gather(bboxes * [height, width, height, width], [1, 0, 3, 2, 1, 2, 3, 0], axis=-1)
    # -> [[left, top], [right, bottom], [left, bottom], [right, top]]
    corners = tf.reshape(corners, [-1, 4, 2])
    # pad 1
    corners = tf.concat([corners, tf.ones_like(corners[:, :, :1])], axis=-1)
    # apply affine transform
    corners_transformed = tf.matmul(corners, affine, transpose_b=True)

    new_left = tf.minimum(corners_transformed[:, 0, 0], corners_transformed[:, 2, 0]) / width
    new_top = tf.minimum(corners_transformed[:, 0, 1], corners_transformed[:, 3, 1]) / height
    new_right = tf.maximum(corners_transformed[:, 1, 0], corners_transformed[:, 3, 0]) / width
    new_bottom = tf.maximum(corners_transformed[:, 1, 1], corners_transformed[:, 2, 1]) / height
    return tf.stack([new_top, new_left, new_bottom, new_right], axis=-1)


def refine_bboxes_labels_single(bboxes, labels):
    bboxes = tf.clip_by_value(bboxes, 0, 1)
    picking_indices = tf.where(tf.not_equal((bboxes[:, 2] - bboxes[:, 0]) * (bboxes[:, 3] - bboxes[:, 1]), 0))
    bboxes, labels = tf.gather_nd(bboxes, picking_indices), tf.gather_nd(labels, picking_indices)
    return bboxes, labels


""" Image Augment """


def random_largest_crop_and_resize_images(images, target_shape, method="bilinear", antialias=False):
    if tf.rank(images) == 3:
        height, width = tf.cast(tf.shape(images)[0], "float32"), tf.cast(tf.shape(images)[1], "float32")
    else:
        height, width = tf.cast(tf.shape(images)[1], "float32"), tf.cast(tf.shape(images)[2], "float32")
    target_height, target_width = tf.cast(target_shape[0], "float32"), tf.cast(target_shape[1], "float32")
    scale = tf.maximum(target_height / height, target_width / width)
    scaled_hh, scaled_ww = tf.cast(tf.math.ceil(height * scale), "int32"), tf.cast(tf.math.ceil(width * scale), "int32")
    images = tf.image.resize(images, [scaled_hh, scaled_ww], method=method, antialias=antialias)

    # print(target_shape)
    crop_hh = tf.cond(scaled_hh > target_shape[0], lambda: tf.random.uniform((), 0, scaled_hh - target_shape[0], dtype="int32"), lambda: 0)
    crop_ww = tf.cond(scaled_ww > target_shape[1], lambda: tf.random.uniform((), 0, scaled_ww - target_shape[1], dtype="int32"), lambda: 0)
    # crop_hh = tf.random.uniform((), 0, tf.maximum(scaled_hh - target_shape[0], 1), dtype='int32')
    # crop_ww = tf.random.uniform((), 0, tf.maximum(scaled_ww - target_shape[1], 1), dtype='int32')
    if tf.rank(images) == 3:
        images = images[crop_hh : crop_hh + target_shape[0], crop_ww : crop_ww + target_shape[1]]
    else:
        images = images[:, crop_hh : crop_hh + target_shape[0], crop_ww : crop_ww + target_shape[1]]
    return images, scale, crop_hh, crop_ww


def get_image_aspect_aware_random_scale_crop(source_shape, target_shape, scale_min=0.1, scale_max=2.0):
    """https://github.com/google/automl/tree/master/efficientdet/dataloader.py#L67"""
    random_scale_factor = tf.random.uniform((), scale_min, scale_max)
    scaled_y, scaled_x = random_scale_factor * target_shape[0], random_scale_factor * target_shape[1]
    height, width = tf.cast(source_shape[0], tf.float32), tf.cast(source_shape[1], tf.float32)
    random_image_scale = tf.minimum(scaled_y / height, scaled_x / width)

    # Select non-zero random offset (x, y) if scaled image is larger than self._output_size.
    height, width = tf.cast(source_shape[0], tf.float32), tf.cast(source_shape[1], tf.float32)
    scaled_height, scaled_width = height * random_image_scale, width * random_image_scale
    offset_y, offset_x = tf.maximum(0.0, scaled_height - target_shape[0]), tf.maximum(0.0, scaled_width - target_shape[1])
    random_offset_y, random_offset_x = offset_y * tf.random.uniform([], 0, 1), offset_x * tf.random.uniform([], 0, 1)
    random_offset_y, random_offset_x = tf.cast(random_offset_y, tf.int32), tf.cast(random_offset_x, tf.int32)
    return random_image_scale, random_offset_y, random_offset_x


def aspect_aware_resize_and_crop_image(image, target_shape, scale=-1, crop_y=0, crop_x=0, letterbox_pad=-1, method="bilinear", antialias=False):
    letterbox_target_shape = (target_shape[0] - letterbox_pad, target_shape[1] - letterbox_pad) if letterbox_pad > 0 else target_shape
    height, width = tf.cast(tf.shape(image)[0], "float32"), tf.cast(tf.shape(image)[1], "float32")
    if scale == -1:
        scale = tf.minimum(letterbox_target_shape[0] / height, letterbox_target_shape[1] / width)
    scaled_hh, scaled_ww = int(height * scale), int(width * scale)
    # image = tf.cast(image, "float32")
    image = tf.image.resize(image, [scaled_hh, scaled_ww], method=method, antialias=antialias)
    image = image[crop_y : crop_y + letterbox_target_shape[0], crop_x : crop_x + letterbox_target_shape[1]]
    cropped_shape = tf.shape(image)

    pad_top, pad_left = ((target_shape[0] - cropped_shape[0]) // 2, (target_shape[1] - cropped_shape[1]) // 2) if letterbox_pad >= 0 else (0, 0)
    image = tf.image.pad_to_bounding_box(image, pad_top, pad_left, target_shape[0], target_shape[1])
    return image, scale, pad_top, pad_left


def random_flip_left_right_with_bboxes(image, bboxes, probability=0.5):
    # For box, left = 1 - right, right = 1 - left
    return tf.cond(
        tf.random.uniform(()) < probability,
        lambda: (tf.image.flip_left_right(image), tf.gather(bboxes, [0, 3, 2, 1], axis=1) * [1, -1, 1, -1] + [0, 1, 0, 1]),
        lambda: (image, bboxes),
    )


def random_hsv(image, hue_delta=0.015, saturation_delta=0.7, brightness_delta=0.4, contrast_delta=0, show_sample=False):
    # augment_hsv https://github.com/WongKinYiu/yolor/blob/main/utils/datasets.py#L941
    if show_sample:
        import matplotlib.pyplot as plt

        fig = plt.figure()
        aa = tf.concat([tf.image.adjust_hue(image, -hue_delta), tf.image.adjust_hue(image, hue_delta)], axis=1)
        bb = tf.concat([tf.image.adjust_saturation(image, 1 - saturation_delta), tf.image.adjust_saturation(image, 1 + saturation_delta)], axis=1)
        cc = tf.concat([tf.image.adjust_brightness(image, -brightness_delta), tf.image.adjust_brightness(image, brightness_delta)], axis=1)
        dd = tf.concat([tf.image.adjust_contrast(image, 1 - contrast_delta), tf.image.adjust_contrast(image, 1 + contrast_delta)], axis=1)
        plt.imshow(tf.concat([aa, bb, cc, dd], axis=0))
        plt.axis("off")
        plt.tight_layout()

    image = tf.image.random_hue(image, hue_delta)
    image = tf.image.random_saturation(image, 1 - saturation_delta, 1 + saturation_delta)
    image = tf.image.random_brightness(image, brightness_delta)
    if contrast_delta > 0:
        image = tf.image.random_contrast(image, 1 - contrast_delta, 1 + contrast_delta)
    return image


""" Mosaic mix """


def mosaic_mix_batch(images, bboxes, labels, split_center_min=0.25, split_center_max=0.75):
    batch_size = tf.shape(images)[0]
    _, hh, ww, _ = images.shape
    split_hh = tf.cast(tf.random.uniform((), split_center_min * hh, split_center_max * hh), "int32")
    split_ww = tf.cast(tf.random.uniform((), split_center_min * ww, split_center_max * ww), "int32")
    # print(split_hh, split_ww)

    # top_left, top_right, bottom_left, bottom_right
    # sub_hh_wws = [[split_hh, split_ww], [split_hh, ww - split_ww], [hh - split_hh, split_ww], [hh - split_hh, ww - split_ww]]
    starts = [[0, 0], [0, split_ww], [split_hh, 0], [split_hh, split_ww]]
    ends = [[split_hh, split_ww], [split_hh, ww], [hh, split_ww], [hh, ww]]
    mixed_images, mixed_bboxes, mixed_labels = [], [], []
    for (top, left), (bottom, right) in zip(starts, ends):
        sub_hh, sub_ww = bottom - top, right - left
        pick_indices = tf.random.shuffle(tf.range(batch_size))
        cur_images = tf.gather(images, pick_indices)
        cur_images, scale, crop_hh, crop_ww = random_largest_crop_and_resize_images(cur_images, [sub_hh, sub_ww])
        mixed_images.append(cur_images)
        # print(f"{cur_images.shape = }, {scale = }")

        cur_bboxes, cur_labels = tf.gather(bboxes, pick_indices), tf.gather(labels, pick_indices)
        target_height, target_width = 1, 1  # Don't scale back to [0, 1]
        scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(hh, ww, target_height, target_width, scale, scale, crop_hh, crop_ww)
        cur_bboxes = cur_bboxes * [scale_hh, scale_ww, scale_hh, scale_ww] - [crop_hh, crop_ww, crop_hh, crop_ww]

        """ Re-fine batch bboxes """
        sub_hh, sub_ww = tf.cast(sub_hh, cur_bboxes.dtype), tf.cast(sub_ww, cur_bboxes.dtype)
        cur_bboxes = tf.clip_by_value(cur_bboxes, 0, [sub_hh, sub_ww, sub_hh, sub_ww])
        mask_cond = tf.not_equal((cur_bboxes[:, :, 2] - cur_bboxes[:, :, 0]) * (cur_bboxes[:, :, 3] - cur_bboxes[:, :, 1]), 0)
        cur_bboxes += [top, left, top, left]
        cur_bboxes = cur_bboxes * tf.expand_dims(tf.cast(mask_cond, cur_bboxes.dtype), -1)
        cur_labels = tf.where(mask_cond, cur_labels, tf.zeros_like(cur_labels))

        mixed_bboxes.append(cur_bboxes)
        mixed_labels.append(cur_labels)

    top_images = tf.concat([mixed_images[0], mixed_images[1]], axis=2)
    bottom_images = tf.concat([mixed_images[2], mixed_images[3]], axis=2)
    mixed_images = tf.concat([top_images, bottom_images], axis=1)
    mixed_images.set_shape([None, hh, ww, 3])
    mixed_bboxes = tf.concat(mixed_bboxes, axis=1) / [hh, ww, hh, ww]
    mixed_labels = tf.concat(mixed_labels, axis=1)
    # print(f"{top_images.shape = }, {bottom_images.shape = }, {mix.shape = }, {mixed_images.shape = }, {mixed_labels.shape = }")
    return mixed_images, (mixed_bboxes, mixed_labels)


""" Dataset """


class RandomProcessImageWithBboxes:
    def __init__(
        self,
        target_shape=(300, 300),
        max_labels_per_image=100,
        random_crop_mode=0,  # 0 for eval mode, (0, 1) for random crop, 1 for random largest crop, > 1 for random scale
        resize_method="bilinear",
        resize_antialias=False,
        color_augment_method="random_hsv",
        magnitude=0,
        num_layers=2,
        use_color_increasing=True,
        **randaug_kwargs,
    ):
        self.max_labels_per_image = max_labels_per_image
        self.target_shape = target_shape if len(target_shape) == 2 else target_shape[:2]
        self.resize_method, self.resize_antialias, self.random_crop_mode, self.magnitude = resize_method, resize_antialias, random_crop_mode, magnitude
        if magnitude > 0:
            from keras_cv_attention_models.imagenet import augment

            if color_augment_method is None:
                self.randaug_wo_pos = lambda image: image
            elif isinstance(color_augment_method, str) and color_augment_method.lower() == "autoaug":
                self.randaug_wo_pos = augment.AutoAugment(augmentation_name="simple")
            elif isinstance(color_augment_method, str) and color_augment_method.lower() == "randaug":
                # TODO: Need to pick color related methods, "Invert" / "Posterize" may not working well here.
                print(">>>> RandAugment: magnitude = %d" % magnitude)
                self.randaug_wo_pos = augment.RandAugment(
                    num_layers=num_layers,
                    magnitude=magnitude,
                    use_cutout=False,
                    use_color_increasing=use_color_increasing,
                    use_positional_related_ops=False,  # Set False to exlude [shear, rotate, translate]
                    **randaug_kwargs,
                )
            elif isinstance(color_augment_method, str):
                self.randaug_wo_pos = random_hsv
            else:
                print(">>>> Using user defined color_augment_method: {}".format(getattr(color_augment_method, "__name__", color_augment_method)))
                self.randaug_wo_pos = color_augment_method

            # RandAugment positional related ops. Including [shear, rotate, translate], Also returns affine transform matrix
            # self.pos_randaug = augment.PositionalRandAugment(num_layers=num_layers, magnitude=magnitude, **randaug_kwargs)

    def __random_crop_and_resize__(self, image, bbox):
        height, width = tf.shape(image)[0], tf.shape(image)[1]

        if self.random_crop_mode > 1:
            scale, crop_hh, crop_ww = get_image_aspect_aware_random_scale_crop((height, width), self.target_shape, scale_max=self.random_crop_mode)
            image, scale, _, _ = aspect_aware_resize_and_crop_image(
                image, self.target_shape, scale, crop_hh, crop_ww, method=self.resize_method, antialias=self.resize_antialias
            )
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(
                height, width, self.target_shape[0], self.target_shape[1], scale, scale, crop_hh, crop_ww
            )
        elif self.random_crop_mode == 1:
            image, scale, crop_hh, crop_ww = random_largest_crop_and_resize_images(image, self.target_shape, self.resize_method, self.resize_antialias)
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(
                height, width, self.target_shape[0], self.target_shape[1], scale, scale, crop_hh, crop_ww
            )
        elif self.random_crop_mode > 0 and self.random_crop_mode < 1:
            image, scale_hh, scale_ww, crop_hh, crop_ww = random_crop_and_resize_image(
                image, self.target_shape, scale=(self.random_crop_mode, 1.0), method=self.resize_method, antialias=self.resize_antialias
            )
        else:
            image, scale, _, _ = aspect_aware_resize_and_crop_image(image, self.target_shape, method=self.resize_method, antialias=self.resize_antialias)
            scale_hh, scale_ww, crop_hh, crop_ww = rerange_scale_offset_to_01(height, width, self.target_shape[0], self.target_shape[1], scale, scale)

        bbox = bbox * [scale_hh, scale_ww, scale_hh, scale_ww] - [crop_hh, crop_ww, crop_hh, crop_ww]
        return image, bbox

    def __call__(self, datapoint):
        image = datapoint["image"]
        objects = datapoint["objects"]
        bbox, label, is_crowd = tf.cast(objects["bbox"], tf.float32), objects["label"], objects.get("is_crowd", None)
        if len(image.shape) < 2:
            image = tf_imread(image)
        if is_crowd is not None:
            is_not_crowd = is_crowd == False
            bbox, label = tf.boolean_mask(bbox, is_not_crowd), tf.boolean_mask(label, is_not_crowd)

        if self.magnitude >= 0:
            image, bbox = random_flip_left_right_with_bboxes(image, bbox)

        image, bbox = self.__random_crop_and_resize__(image, bbox)
        bbox, label = refine_bboxes_labels_single(bbox, label)

        if self.magnitude > 0:
            image.set_shape([*self.target_shape[:2], 3])
            image = self.randaug_wo_pos(image)

        should_pad = self.max_labels_per_image - tf.shape(bbox)[0]
        # label starts from 0 -> person, add 1 here, differs from padded values
        bbox, label = tf.cond(
            should_pad > 0,
            lambda: (tf.pad(bbox, [[0, should_pad], [0, 0]]), tf.pad(label + 1, [[0, should_pad]])),
            lambda: (bbox[: self.max_labels_per_image], label[: self.max_labels_per_image] + 1),
        )

        image = tf.cast(image, tf.float32)
        image.set_shape([*self.target_shape[:2], 3])
        return image, (bbox, label)


class PositionalRandAugmentWithBboxes:
    def __init__(self, magnitude=0, num_layers=2, max_labels_per_image=100, positional_augment_methods="rts", **randaug_kwargs):
        from keras_cv_attention_models.imagenet import augment

        self.pos_randaug = augment.PositionalRandAugment(
            num_layers=num_layers, magnitude=magnitude, positional_augment_methods=positional_augment_methods, **randaug_kwargs
        )
        self.max_labels_per_image = max_labels_per_image

    def __call_single__(self, inputs):
        image, bbox, label = inputs
        image, affine_matrix = self.pos_randaug(image)
        pick = tf.where(label != 0)
        bbox, label = tf.gather_nd(bbox, pick), tf.gather_nd(label, pick)
        bbox = bboxes_apply_affine(bbox, affine_matrix, input_shape=image.shape)
        bbox, label = refine_bboxes_labels_single(bbox, label)

        should_pad = self.max_labels_per_image - tf.shape(bbox)[0]
        bbox, label = tf.cond(
            should_pad > 0,
            lambda: (tf.pad(bbox, [[0, should_pad], [0, 0]]), tf.pad(label, [[0, should_pad]])),
            lambda: (bbox[: self.max_labels_per_image], label[: self.max_labels_per_image]),
        )
        return image, bbox, label

    def __call__(self, xx, yy):
        image, bbox, label = tf.map_fn(self.__call_single__, (xx, yy[0], yy[1]))
        return image, (bbox, label)


def to_one_hot_with_class_mark(anchor_bboxes_with_label, num_classes=80):
    # dest_boxes, anchor_classes = anchor_bboxes_with_label[:, :4], anchor_bboxes_with_label[:, -1]
    dest_boxes, anchor_classes = tf.split(anchor_bboxes_with_label, [4, 1], axis=-1)
    one_hot_labels = tf.one_hot(tf.cast(anchor_classes[..., 0], "int32") - 1, num_classes)  # [1, 81] -> [0, 80]
    # Mark iou < ignore_threshold as 0, ignore_threshold < iou < overlap_threshold as -1, iou > overlap_threshold as 1
    marks = tf.where(anchor_classes > 0, tf.ones_like(anchor_classes), anchor_classes)
    # marks = tf.expand_dims(marks, -1)
    one_hot_labels, marks = tf.cast(one_hot_labels, dest_boxes.dtype), tf.cast(marks, dest_boxes.dtype)
    return tf.concat([dest_boxes, one_hot_labels, marks], axis=-1)


def __bboxes_labels_batch_func__(bboxes, labels, anchors, empty_label, num_classes=80):
    bbox_labels = tf.concat([bboxes, tf.cast(tf.expand_dims(labels, -1), bboxes.dtype)], axis=-1)
    bbox_process = lambda xx: tf.cond(
        tf.reduce_any(xx[:, -1] > 0),  # If contains any valid bbox and label
        lambda: to_one_hot_with_class_mark(anchors_func.assign_anchor_classes_by_iou_with_bboxes(xx, anchors), num_classes),
        lambda: empty_label,
    )
    return tf.map_fn(bbox_process, bbox_labels)


def __yolor_bboxes_labels_batch_func__(bboxes, labels, anchor_ratios, feature_sizes, empty_label, num_classes=80):
    bbox_labels = tf.concat([bboxes, tf.cast(tf.expand_dims(labels, -1), bboxes.dtype)], axis=-1)
    bbox_process = lambda xx: tf.cond(
        tf.reduce_any(xx[:, -1] > 0),  # If contains any valid bbox and label
        lambda: to_one_hot_with_class_mark(anchors_func.yolor_assign_anchors(xx, anchor_ratios, feature_sizes), num_classes),
        lambda: empty_label,
    )
    return tf.map_fn(bbox_process, bbox_labels)


def detection_dataset_from_custom_json(data_path, with_info=False):
    import json

    with open(data_path, "r") as ff:
        aa = json.load(ff)
    test_key = "validation" if "validation" in aa else "test"
    train, test, info = aa["train"], aa[test_key], aa.get("info", {})
    total_images, num_classes = len(train), info.get("num_classes", 0)
    if num_classes <= 0:
        num_classes = max([max([int(jj) for jj in ii["objects"]["label"]]) for ii in train]) + 1
        print(">>>> Using max value from train as num_classes:", num_classes)

    if "base_path" in info and len(info["base_path"]) > 0:
        base_path = os.path.expanduser(info["base_path"])
        for ii in train:
            ii["image"] = os.path.join(base_path, ii["image"])
        for ii in test:
            ii["image"] = os.path.join(base_path, ii["image"])

    objects_signature = {"bbox": tf.TensorSpec(shape=(None, 4), dtype=tf.float32), "label": tf.TensorSpec(shape=(None,), dtype=tf.int64)}
    output_signature = {"image": tf.TensorSpec(shape=(), dtype=tf.string), "objects": objects_signature}
    train_ds = tf.data.Dataset.from_generator(lambda: (ii for ii in train), output_signature=output_signature)
    test_ds = tf.data.Dataset.from_generator(lambda: (ii for ii in test), output_signature=output_signature)

    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    train_ds = train_ds.apply(tf.data.experimental.assert_cardinality(len(train))).with_options(options)
    test_ds = test_ds.apply(tf.data.experimental.assert_cardinality(len(test))).with_options(options)
    dataset = {"train": train_ds, test_key: test_ds}
    return (dataset, total_images, num_classes) if with_info else dataset


def init_dataset(
    data_name="coco/2017",  # dataset params
    input_shape=(256, 256),
    batch_size=64,
    buffer_size=1000,
    info_only=False,
    max_labels_per_image=100,
    anchors_mode="efficientdet",
    anchor_pyramid_levels=[3, 7],
    anchor_scale=4,  # For efficientdet anchors only
    aspect_ratios=(1, 2, 0.5),  # For efficientdet anchors only
    num_scales=3,  # For efficientdet anchors only
    rescale_mode="torch",  # rescale mode, ["tf", "torch"], or specific `(mean, std)` like `(128.0, 128.0)`
    resize_method="bilinear",  # ["bilinear", "bicubic"]
    resize_antialias=False,
    random_crop_mode=1.0,  # [augment] 0 for eval mode, (0, 1) for random crop, 1 for random largest crop, > 1 for random scale
    mosaic_mix_prob=0.0,  # [augment] 0 for disable, > 0 for enable
    color_augment_method="random_hsv",  # [augment] one of ["random_hsv", "autoaug", "randaug"], or totally custom one like `lambda image: image`
    positional_augment_methods="rts",  # [augment] Positional augment method besides scale, combine of r: rotate, t: transplate, s: shear, x: scale_x + scale_y
    magnitude=0,
    num_layers=2,
    seed=None,
    **augment_kwargs,  # Too many...
):
    import tensorflow_datasets as tfds

    is_tpu = True if len(tf.config.list_logical_devices("TPU")) > 0 else False  # Set True for try_gcs and drop_remainder
    try_gcs, drop_remainder = is_tpu, is_tpu

    if data_name.endswith(".json"):
        dataset, total_images, num_classes = detection_dataset_from_custom_json(data_name, with_info=True)
    else:
        dataset, info = tfds.load(data_name, with_info=True, try_gcs=try_gcs)
        num_classes = info.features["objects"]["label"].num_classes
        total_images = info.splits["train"].num_examples
    steps_per_epoch = int(tf.math.ceil(total_images / float(batch_size)))
    if info_only:
        return total_images, num_classes, steps_per_epoch

    AUTOTUNE = tf.data.AUTOTUNE

    train_process = RandomProcessImageWithBboxes(
        target_shape=input_shape,
        max_labels_per_image=max_labels_per_image,
        random_crop_mode=random_crop_mode,
        resize_method=resize_method,
        resize_antialias=resize_antialias,
        color_augment_method=color_augment_method,
        magnitude=magnitude,
        num_layers=num_layers,
        **augment_kwargs,
    )
    train_dataset = dataset["train"].shuffle(buffer_size, seed=seed).map(train_process).batch(batch_size, drop_remainder=drop_remainder)
    # return train_dataset

    if mosaic_mix_prob > 0:
        mosaic_mix = lambda xx, yy: tf.cond(
            tf.random.uniform(()) > mosaic_mix_prob,
            lambda: (xx, yy),
            lambda: mosaic_mix_batch(xx, yy[0], yy[1]),
        )
        train_dataset = train_dataset.map(mosaic_mix, num_parallel_calls=AUTOTUNE)
    # return train_dataset

    if magnitude > 0 and positional_augment_methods is not None and len(positional_augment_methods) != 0:
        # Apply randaug rotate / shear / transform after mosaic mix
        max_labels_per_image = (max_labels_per_image * 4) if mosaic_mix_prob > 0 else max_labels_per_image
        pos_aug = PositionalRandAugmentWithBboxes(magnitude, num_layers, max_labels_per_image, positional_augment_methods, **augment_kwargs)
        print(">>>> positional augment methods:", pos_aug.pos_randaug.available_ops)
        train_dataset = train_dataset.map(pos_aug, num_parallel_calls=AUTOTUNE)

    if anchors_mode == anchors_func.ANCHOR_FREE_MODE or anchors_mode == anchors_func.YOLOV8_MODE:  # == "anchor_free"
        # Don't need anchors here, anchor assigning is after getting model predictions.
        bbox_process = lambda bb: to_one_hot_with_class_mark(tf.concat([bb[0], tf.cast(tf.expand_dims(bb[1], -1), bb[0].dtype)], axis=-1), num_classes)
    elif anchors_mode == anchors_func.YOLOR_MODE:  # == "yolor":
        anchor_ratios, feature_sizes = anchors_func.get_yolor_anchors(input_shape[:2], anchor_pyramid_levels, is_for_training=True)
        total_anchors = tf.cast(anchor_ratios.shape[1] * tf.reduce_sum(feature_sizes[:, 0] * feature_sizes[:, 1]), tf.int32)
        empty_label = tf.zeros([total_anchors, 4 + num_classes + 1])  # All 0
        bbox_process = lambda bb: __yolor_bboxes_labels_batch_func__(bb[0], bb[1], anchor_ratios, feature_sizes, empty_label, num_classes)
    else:
        # grid_zero_start = True if anchor_grid_zero_start == "auto" else anchor_grid_zero_start
        grid_zero_start = False  # Use this till meet some others new
        anchors = anchors_func.get_anchors(input_shape[:2], anchor_pyramid_levels, aspect_ratios, num_scales, anchor_scale, grid_zero_start)
        num_anchors = anchors.shape[0]
        empty_label = tf.zeros([num_anchors, 4 + num_classes + 1])  # All 0
        bbox_process = lambda bb: __bboxes_labels_batch_func__(bb[0], bb[1], anchors, empty_label, num_classes)

    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    rescaling = lambda xx: (xx - mean) / std
    train_dataset = train_dataset.map(lambda xx, yy: (rescaling(xx), bbox_process(yy)), num_parallel_calls=AUTOTUNE)
    train_dataset = train_dataset.prefetch(buffer_size=AUTOTUNE)
    # return train_dataset

    """ Test dataset """
    test_dataset = dataset.get("validation", dataset.get("test", None))
    if test_dataset is not None:
        test_process = RandomProcessImageWithBboxes(target_shape=input_shape, resize_method=resize_method, resize_antialias=resize_antialias, magnitude=-1)
        test_dataset = test_dataset.map(test_process).batch(batch_size, drop_remainder=drop_remainder).map(lambda xx, yy: (rescaling(xx), bbox_process(yy)))

    return train_dataset, test_dataset, total_images, num_classes, steps_per_epoch


""" Show """


def show_batch_sample(
    dataset, rescale_mode="torch", rows=-1, label_font_size=8, base_size=3, anchors_mode="efficientdet", indices_2_labels=None, **anchor_kwargs
):
    import matplotlib.pyplot as plt
    from keras_cv_attention_models.plot_func import get_plot_cols_rows, show_image_with_bboxes

    if isinstance(dataset, (list, tuple)):
        images, labels = dataset
    else:
        images, labels = dataset.as_numpy_iterator().next()
    mean, std = init_mean_std_by_rescale_mode(rescale_mode)
    images = (images * std + mean) / 255
    if anchors_mode == anchors_func.YOLOR_MODE:
        pyramid_levels = anchors_func.get_pyramid_levels_by_anchors(images.shape[1:-1], labels.shape[1])
        anchors = anchors_func.get_yolor_anchors(images.shape[1:-1], pyramid_levels=pyramid_levels, is_for_training=False)
    elif not anchors_mode == anchors_func.ANCHOR_FREE_MODE:
        pyramid_levels = anchors_func.get_pyramid_levels_by_anchors(images.shape[1:-1], labels.shape[1])
        anchors = anchors_func.get_anchors(images.shape[1:-1], pyramid_levels, **anchor_kwargs)

    cols, rows = get_plot_cols_rows(len(images), rows, ceil_mode=True)
    fig, axes = plt.subplots(rows, cols, figsize=(base_size * cols, base_size * rows))
    axes = axes.flatten()

    for ax, image, label in zip(axes, images, labels):
        if label.shape[-1] == 5:
            pick = label[:, -1] >= 0
            valid_preds = label[pick]
        else:
            pick = label[:, -1] == 1
            valid_preds = label[pick]
            valid_label = tf.cast(tf.argmax(valid_preds[:, 4:-1], axis=-1), valid_preds.dtype)
            valid_preds = tf.concat([valid_preds[:, :4], tf.expand_dims(valid_label, -1)], axis=-1)

        if anchors_mode == anchors_func.YOLOR_MODE:
            valid_anchors = anchors[pick]
            decoded_centers = (valid_preds[:, :2] + 0.5) * valid_anchors[:, 4:] + valid_anchors[:, :2]
            decoded_hw = valid_preds[:, 2:4] * valid_anchors[:, 4:]
            decoded_corner = anchors_func.center_yxhw_to_corners_nd(tf.concat([decoded_centers, decoded_hw], axis=-1))
            valid_preds = tf.concat([decoded_corner, valid_preds[:, -1:]], axis=-1)
        elif not anchors_mode == anchors_func.ANCHOR_FREE_MODE:
            valid_anchors = anchors[pick]
            valid_preds = anchors_func.decode_bboxes(valid_preds, valid_anchors)
        show_image_with_bboxes(image, valid_preds[:, :4], valid_preds[:, -1], ax=ax, label_font_size=label_font_size, indices_2_labels=indices_2_labels)
    fig.tight_layout()
    plt.show()
    return fig



from keras_cv_attention_models.mobilevit.mobilevit import MobileViT
from keras_cv_attention_models.models import register_model


def MobileViT_V2(
    num_blocks=[1, 2, 3, 5, 4],
    out_channels=[64, 128, 256, 384, 512],
    attn_channels=0.5,  # Can be a list matching out_channels, or a float number for expansion ratio of out_channels
    expand_ratio=2,
    stem_width=32,
    resize_first=True,  # False for V1, True for V2
    use_depthwise=True,  # False for V1, True for V2
    use_fusion=False,  # True for V1, False for V2
    num_norm_groups=1,  # -1 or 0 for V1 using layer_norm, or 1 for V2 using group_norm
    use_linear_attention=True,  # False for V1, True for V2
    output_num_features=0,
    model_name="mobilevit_v2",
    **kwargs,
):
    kwargs.pop("kwargs", None)
    return MobileViT(**locals(), **kwargs)


def get_mobilevit_v2_width(multiplier=1.0):
    return int(32 * multiplier), [int(ii * multiplier) for ii in [64, 128, 256, 384, 512]]  # stem_width, out_channels


@register_model
def MobileViT_V2_050(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(0.5)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_050", **kwargs)


@register_model
def MobileViT_V2_075(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(0.75)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_075", **kwargs)


@register_model
def MobileViT_V2_100(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_100", **kwargs)


@register_model
def MobileViT_V2_125(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(1.25)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_125", **kwargs)


@register_model
def MobileViT_V2_150(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(1.5)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_150", **kwargs)


@register_model
def MobileViT_V2_175(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(1.75)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_175", **kwargs)


@register_model
def MobileViT_V2_200(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_width, out_channels = get_mobilevit_v2_width(2.0)
    return MobileViT_V2(**locals(), model_name="mobilevit_v2_200", **kwargs)



from keras_cv_attention_models import backend

from keras_cv_attention_models.version import __version__
from keras_cv_attention_models import plot_func
from keras_cv_attention_models import attention_layers
from keras_cv_attention_models import beit
from keras_cv_attention_models.beit import flexivit
from keras_cv_attention_models.beit import eva
from keras_cv_attention_models.beit import eva02
from keras_cv_attention_models.beit import dinov2
from keras_cv_attention_models.beit import meta_transformer
from keras_cv_attention_models.beit import vit
from keras_cv_attention_models import botnet
from keras_cv_attention_models import caformer
from keras_cv_attention_models import coat
from keras_cv_attention_models import coatnet
from keras_cv_attention_models import convnext
from keras_cv_attention_models import cotnet
from keras_cv_attention_models import cmt
from keras_cv_attention_models import davit
from keras_cv_attention_models import efficientnet
from keras_cv_attention_models import edgenext
from keras_cv_attention_models import efficientformer
from keras_cv_attention_models import fasternet
from keras_cv_attention_models import gcvit
from keras_cv_attention_models import ghostnet
from keras_cv_attention_models import ghostnet as ghostnetv2  # Will be removed
from keras_cv_attention_models import gpt2
from keras_cv_attention_models import llama2
from keras_cv_attention_models import halonet
from keras_cv_attention_models import hiera
from keras_cv_attention_models import hornet
from keras_cv_attention_models import iformer
from keras_cv_attention_models import levit
from keras_cv_attention_models import mlp_family
from keras_cv_attention_models.mlp_family import mlp_mixer
from keras_cv_attention_models.mlp_family import res_mlp
from keras_cv_attention_models.mlp_family import gated_mlp
from keras_cv_attention_models.mlp_family import wave_mlp
from keras_cv_attention_models.mobilenetv3_family import fbnetv3
from keras_cv_attention_models.mobilenetv3_family import lcnet
from keras_cv_attention_models.mobilenetv3_family import mobilenetv3
from keras_cv_attention_models.mobilenetv3_family import tinynet
from keras_cv_attention_models import efficientvit
from keras_cv_attention_models.efficientvit import efficientvit_b
from keras_cv_attention_models.efficientvit import efficientvit_m
from keras_cv_attention_models import inceptionnext
from keras_cv_attention_models import maxvit
from keras_cv_attention_models import mobilevit
from keras_cv_attention_models import moganet
from keras_cv_attention_models import nat
from keras_cv_attention_models.nat import dinat
from keras_cv_attention_models import pvt
from keras_cv_attention_models import repvit
from keras_cv_attention_models import tinyvit
from keras_cv_attention_models import resnest
from keras_cv_attention_models import resnet_family
from keras_cv_attention_models.resnet_family import resnext
from keras_cv_attention_models.resnet_family import resnet_quad
from keras_cv_attention_models.resnet_family import resnet_deep
from keras_cv_attention_models.resnet_family import regnet
from keras_cv_attention_models import gpvit
from keras_cv_attention_models import swin_transformer_v2
from keras_cv_attention_models import uniformer
from keras_cv_attention_models import fastervit
from keras_cv_attention_models import fastvit
from keras_cv_attention_models import vanillanet
from keras_cv_attention_models import download_and_load
from keras_cv_attention_models import imagenet
from keras_cv_attention_models import test_images
from keras_cv_attention_models import model_surgery
from keras_cv_attention_models import efficientdet
from keras_cv_attention_models import yolox
from keras_cv_attention_models import yolor
from keras_cv_attention_models import yolov7
from keras_cv_attention_models import yolov8
from keras_cv_attention_models.yolov8 import yolo_nas
from keras_cv_attention_models import coco
from keras_cv_attention_models import clip
from keras_cv_attention_models.clip import tokenizer
from keras_cv_attention_models import stable_diffusion

if backend.is_tensorflow_backend:
    from keras_cv_attention_models import nfnets
    from keras_cv_attention_models import volo
    from keras_cv_attention_models import visualizing



import os
import keras_cv_attention_models as __package__


def register_model(model_func):
    if not hasattr(__package__.models, model_func.__name__):
        setattr(__package__.models, model_func.__name__, model_func)
    return model_func


class FakeModelWrapper:
    def __init__(self, models, name="model"):
        self.models = models if isinstance(models, (list, tuple)) else [models]
        self.name = name

    def cuda(self):
        """Torch function"""
        self.models = [model.cuda() for model in self.models]
        return self

    def cpu(self):
        """Torch function"""
        self.models = [model.cpu() for model in self.models]
        return self

    def float(self):
        """Torch function"""
        self.models = [model.float() for model in self.models]
        return self

    def half(self):
        """Torch function"""
        self.models = [model.half() for model in self.models]
        return self

    def to(self, *args):
        """Torch function"""
        self.models = [model.to(*args) for model in self.models]
        return self

    def _save_load_file_path_rule_(self, file_path=None):
        file_path = self.name if file_path is None else file_path
        suffix = os.path.splitext(file_path)[1]
        if suffix in [".h5", ".keras", ".pt", ".pth"]:
            file_path = os.path.splitext(file_path)[0]
            save_path_rule = lambda model_name: file_path + "_" + model_name + suffix
        else:  # Regard as directory
            if not os.path.exists(file_path):
                os.makedirs(file_path, exist_ok=True)
            save_path_rule = lambda model_name: os.path.join(file_path, model_name + ".h5")
        return save_path_rule

    def save(self, file_path=None):
        """file_path: if suffix in [".h5", ".keras", ".pt", ".pth"], will save as {file_path}_{model_name}.{suffix},
        or will regard as directory, and save to {file_path}/{model_name}.h5
        """
        save_path_rule = self._save_load_file_path_rule_(file_path)
        for model in self.models:
            cur_save_path = save_path_rule(model.name)
            print(">>>> Saving {} to {}".format(model.name, cur_save_path))
            model.save(cur_save_path)

    def save_weights(self, file_path=None):
        """file_path: if suffix in [".h5", ".keras", ".pt", ".pth"], will save as {file_path}_{model_name}.{suffix},
        or will regard as directory, and save to {file_path}/{model_name}.h5
        """
        save_path_rule = self._save_load_file_path_rule_(file_path)
        for model in self.models:
            cur_save_path = save_path_rule(model.name)
            print(">>>> Saving {} weights to {}".format(model.name, cur_save_path))
            model.save_weights(cur_save_path)

    def load_weights(self, file_path=None):
        """file_path: if suffix in [".h5", ".keras", ".pt", ".pth"], will load from {file_path}_{model_name}.{suffix},
        or will regard as directory, and load from {file_path}/{model_name}.h5
        """
        save_path_rule = self._save_load_file_path_rule_(file_path)
        for model in self.models:
            cur_save_path = save_path_rule(model.name)
            print(">>>> Loading {} from {}".format(model.name, cur_save_path))
            model.load_weights(cur_save_path)



from keras_cv_attention_models.cotnet.cotnet import CotNet, CotNet50, CotNet101, CotNetSE50D, CotNetSE101D, CotNetSE152D, cot_attention


__head_doc__ = """
Keras implementation of [Github JDAI-CV/CoTNet](https://github.com/JDAI-CV/CoTNet).
Paper [PDF 2107.12292 Contextual Transformer Networks for Visual Recognition](https://arxiv.org/pdf/2107.12292.pdf).
"""

__tail_doc__ = """  bn_after_attn: boolean value if add `batchnorm + activation` layers after attention layer. Default `True`.
  input_shape: it should have exactly 3 inputs channels, like `(224, 224, 3)`.
  num_classes: number of classes to classify images into. Set `0` to exclude top layers.
  activation: activation used in whole model, default `relu`.
  drop_connect_rate: is used for [Deep Networks with Stochastic Depth](https://arxiv.org/abs/1603.09382).
      Can be value like `0.2`, indicates the drop probability linearly changes from `0 --> 0.2` for `top --> bottom` layers.
      A higher value means a higher probability will drop the deep branch.
      or `0` to disable (default).
  classifier_activation: A `str` or callable. The activation function to use on the "top" layer if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer.
      Default is `softmax`.
  pretrained: one of `None` (random initialization) or 'imagenet' (pre-training on ImageNet).
  **kwargs: other parameters from `AotNet` if not conflict.

Returns:
    A `keras.Model` instance.
"""

CotNet.__doc__ = __head_doc__ + """
Args:
  num_blocks: model structure parameter, number of blocks in each stack.
  strides: model structure parameter, strides value for the first block in each stack.
      `[2, 2, 2, 2]` for `CotNetSExxD` models, `[1, 2, 2, 2]` for others.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model        | Params | FLOPs  | Input | Top1 Acc |
  | ------------ |:------:| ------ | ----- |:--------:|
  | CotNet50     | 22.2M  | 3.25G  | 224   |   81.3   |
  | CotNeXt50    | 30.1M  | 4.3G   | 224   |   82.1   |
  | CotNetSE50D  | 23.1M  | 4.05G  | 224   |   81.6   |
  | CotNet101    | 38.3M  | 6.07G  | 224   |   82.8   |
  | CotNeXt101   | 53.4M  | 8.2G   | 224   |   83.2   |
  | CotNetSE101D | 40.9M  | 8.44G  | 224   |   83.2   |
  | CotNetSE152D | 55.8M  | 12.22G | 224   |   84.0   |
  | CotNetSE152D | 55.8M  | 24.92G | 320   |   84.6   |
"""

CotNet50.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

CotNet101.__doc__ = CotNet50.__doc__
CotNetSE50D.__doc__ = CotNet50.__doc__
CotNetSE101D.__doc__ = CotNet50.__doc__
CotNetSE152D.__doc__ = CotNet50.__doc__

cot_attention.__doc__ = __head_doc__ + """
Contextual transformer. Callable function, NOT defined as a layer.
It's using `GroupNormalization` / grouped `Conv2D` / `extract_patches` and other strategies.

Args:
  inputs: input tensor.
  kernel_size: kernel size in this full process.
  activation: activation used after `BatchNormalization`.

Examples:

>>> from keras_cv_attention_models import attention_layers
>>> inputs = keras.layers.Input([28, 28, 192])
>>> nn = attention_layers.cot_attention(inputs, kernel_size=3)
>>> dd = keras.models.Model(inputs, nn)
>>> dd.summary()
>>> dd.output_shape
(None, 28, 28, 192)

>>> {ii.name: ii.shape for ii in dd.weights}
{'key_conv/kernel:0': TensorShape([3, 3, 48, 192]),
 'key_bn/gamma:0': TensorShape([192]),
 'key_bn/beta:0': TensorShape([192]),
 'key_bn/moving_mean:0': TensorShape([192]),
 'key_bn/moving_variance:0': TensorShape([192]),
 'embed_ww_1_conv/kernel:0': TensorShape([1, 1, 384, 96]),
 'embed_ww_1_bn/gamma:0': TensorShape([96]),
 'embed_ww_1_bn/beta:0': TensorShape([96]),
 'embed_ww_1_bn/moving_mean:0': TensorShape([96]),
 'embed_ww_1_bn/moving_variance:0': TensorShape([96]),
 'embed_1_conv/kernel:0': TensorShape([1, 1, 192, 192]),
 'embed_ww_2_conv/kernel:0': TensorShape([1, 1, 96, 216]),
 'embed_ww_2_conv/bias:0': TensorShape([216]),
 'embed_1_bn/gamma:0': TensorShape([192]),
 'embed_1_bn/beta:0': TensorShape([192]),
 'embed_1_bn/moving_mean:0': TensorShape([192]),
 'embed_1_bn/moving_variance:0': TensorShape([192]),
 'embed_ww_group_norm/gamma:0': TensorShape([216]),
 'embed_ww_group_norm/beta:0': TensorShape([216]),
 'embed_2_bn/gamma:0': TensorShape([192]),
 'embed_2_bn/beta:0': TensorShape([192]),
 'embed_2_bn/moving_mean:0': TensorShape([192]),
 'embed_2_bn/moving_variance:0': TensorShape([192]),
 'attn_se_1_conv/kernel:0': TensorShape([1, 1, 192, 96]),
 'attn_se_1_conv/bias:0': TensorShape([96]),
 'attn_se_bn/gamma:0': TensorShape([96]),
 'attn_se_bn/beta:0': TensorShape([96]),
 'attn_se_bn/moving_mean:0': TensorShape([96]),
 'attn_se_bn/moving_variance:0': TensorShape([96]),
 'attn_se_2_conv/kernel:0': TensorShape([1, 1, 96, 384]),
 'attn_se_2_conv/bias:0': TensorShape([384])}
"""



from keras_cv_attention_models.gpt2.gpt2 import (
    GPT2,
    GPT2_Base,
    GPT2_Medium,
    GPT2_Large,
    GPT2_XLarge,
    RunPrediction,
    PositionalIndex,
    CausalMask,
    load_weights_from_huggingface,
)

__head_doc__ = """
Keras implementation of [Github openai/gpt-2](https://github.com/openai/gpt-2).
Paper [Language Models are Unsupervised Multitask Learners](https://d4mucfpksywv.cloudfront.net/better-language-models/language-models.pdf).
"""

__tail_doc__ = """  vocab_size: model vocab size.
  max_block_size: number of tokens generated in each sample.
  include_top: boolena value if including output Dense head layer. Set false to exclude the head layer.
  dropout: float value for drop out rate for Embedding layer and attention blocks.
  activation: activation used in whole model, default `gelu/app`.
  pretrained: None or one of ["webtext", "huggingface"].
      - if "webtext", will try to download and load ported weights if available.
      - if "huggingface", will try converting and loading weights from huggingface `transformers` pacakge.
      - if None, will initialize model with ranbdom weights.

Returns:
    A `keras.Model` instance.
"""

GPT2.__doc__ = __head_doc__ + """
Args:
  num_blocks: num of `attention_mlp_block`s.
  embedding_size: `attention_mlp_block` block embedding size.
  num_heads: num of heads.
  block_use_bias: boolean value if using bias for `attention_mlp_block` Dense layers.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model       | Params  | FLOPs   | vocab_size | LAMBADA PPL |
  | ------------| ------- | ------- | ---------- | ----------- |
  | GPT2_Base   | 163.04M | 146.42G | 50257      | 35.13       |
  | GPT2_Medium | 406.29M | 415.07G | 50257      | 15.60       |
  | GPT2_Large  | 838.36M | 890.28G | 50257      | 10.87       |
  | GPT2_XLarge | 1.638B  | 1758.3G | 50257      | 8.63        |
"""

GPT2_Base.__doc__ = __head_doc__ + """
Args:
""" + __tail_doc__

GPT2_Medium.__doc__ = GPT2_Base.__doc__
GPT2_Large.__doc__ = GPT2_Base.__doc__
GPT2_XLarge.__doc__ = GPT2_Base.__doc__



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    ChannelAffine,
    MultiHeadRelativePositionalEmbedding,
    activation_by_name,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    layer_norm,
    mhsa_with_multi_head_relative_position_embedding,
    mlp_block,
    output_block,
    se_module,
    window_attention,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "gcvit_xx_tiny": {"imagenet": {224: "ff516a5a1d3dfdda0c0b2e0051206c00"}},
    "gcvit_x_tiny": {"imagenet": {224: "723155237e083716bb3df904c80711c4"}},
    "gcvit_tiny": {"imagenet": {224: "0e6ecf576b649f7077f4f2f8122b420e"}},
    "gcvit_small": {"imagenet": {224: "ac0cfb4240ae85a40a88691c2329edab"}},
    "gcvit_base": {"imagenet": {224: "ef6e4015239f68dcabbb8ae9cb799d76"}},
}


def gcvit_block(inputs, window_size, num_heads=4, global_query=None, mlp_ratio=4, layer_scale=0, drop_rate=0, activation="gelu", name=""):
    # print(global_query)
    input_channel = inputs.shape[-1]  # Channels_last only
    attn = layer_norm(inputs, axis=-1, name=name + "attn_")
    attention_block = lambda inputs, num_heads, name: mhsa_with_multi_head_relative_position_embedding(
        inputs, num_heads=num_heads, global_query=global_query, qkv_bias=True, out_bias=True, data_format="channels_last", name=name
    )
    attn = window_attention(attn, window_size=window_size, num_heads=num_heads, attention_block=attention_block, name=name + "window_mhsa_")
    attn = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "1_gamma")(attn) if layer_scale >= 0 else attn
    attn = drop_block(attn, drop_rate=drop_rate, name=name + "attn_")
    attn_out = layers.Add(name=name + "attn_out")([inputs, attn])

    mlp = layer_norm(attn_out, axis=-1, name=name + "mlp_")
    mlp = mlp_block(mlp, int(input_channel * mlp_ratio), use_conv=False, activation=activation, name=name + "mlp_")
    mlp = ChannelAffine(use_bias=False, weight_init_value=layer_scale, axis=-1, name=name + "2_gamma")(mlp) if layer_scale >= 0 else mlp
    mlp = drop_block(mlp, drop_rate=drop_rate, name=name + "mlp_")
    return layers.Add(name=name + "output")([attn_out, mlp])


def to_global_query(inputs, window_ratio, num_heads=4, activation="gelu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    query = inputs
    num_window = 1
    if window_ratio == 1:
        query = extract_feature(query, strides=1, activation=activation, name=name + "down1_")
    else:
        while num_window < window_ratio:
            num_window *= 2
            query = extract_feature(query, strides=2, activation=activation, name=name + "down{}_".format(num_window))

    # print(f"{inputs.shape = }, {query.shape = }, {num_window = }, {window_ratio = }")
    if image_data_format() == "channels_last":
        query = functional.reshape(query, [-1, query.shape[1] * query.shape[2], num_heads, input_channel // num_heads])
        query = functional.transpose(query, [0, 2, 1, 3])  # [batch, num_heads, hh * ww, key_dims]
    else:
        query = functional.reshape(query, [-1, num_heads, input_channel // num_heads, query.shape[2] * query.shape[3]])
        query = functional.transpose(query, [0, 1, 3, 2])  # also [batch, num_heads, hh * ww, key_dims]
    query = functional.repeat(query, num_window * num_window, axis=0)
    # print(f"{query.shape = }")
    return query


def down_sample(inputs, out_channels=-1, activation="gelu", name=""):
    out_channels = out_channels if out_channels > 0 else inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = layer_norm(inputs, name=name + "down_1_")
    nn = extract_feature(nn, strides=1, activation=activation, name=name + "down_")
    nn = conv2d_no_bias(nn, out_channels, kernel_size=3, strides=2, padding="same", name=name + "down_")
    nn = layer_norm(nn, name=name + "down_2_")
    return nn


def extract_feature(inputs, strides=2, activation="gelu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = depthwise_conv2d_no_bias(inputs, kernel_size=3, padding="same", name=name + "extract_")
    nn = activation_by_name(nn, activation=activation, name=name + "extract_")
    nn = se_module(nn, divisor=1, use_bias=False, activation=activation, use_conv=False, name=name + "extract_se_")
    nn = conv2d_no_bias(nn, input_channel, kernel_size=1, name=name + "extract_")
    nn = inputs + nn
    return layers.MaxPool2D(pool_size=3, strides=strides, padding="same", name=name + "extract_maxpool")(nn) if strides > 1 else nn
    # if strides > 1:
    #     nn = tf.pad(nn, [[0, 0], [1, 1], [1, 1], [0, 0]])
    #     nn = layers.MaxPool2D(pool_size=3, strides=strides, padding="valid", name=name + "extract_maxpool")(nn)
    # return nn


def GCViT(
    num_blocks=[2, 2, 6, 2],
    num_heads=[2, 4, 8, 16],
    # window_size=[7, 7, 14, 7],
    window_ratios=[8, 4, 1, 1],
    embed_dim=64,
    mlp_ratio=3,
    layer_scale=-1,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="gelu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    dropout=0,
    pretrained=None,
    model_name="gcvit",
    kwargs=None,
):
    """Patch stem"""
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    nn = conv2d_no_bias(inputs, embed_dim, kernel_size=3, strides=2, use_bias=True, padding="same", name="stem_conv")
    nn = down_sample(nn, name="stem_")
    height_axis, width_axis = (1, 2) if image_data_format() == "channels_last" else (2, 3)

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    num_stacks = len(num_blocks)
    for stack_id, (num_block, num_head, window_ratio) in enumerate(zip(num_blocks, num_heads, window_ratios)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = down_sample(nn, out_channels=nn.shape[-1 if image_data_format() == "channels_last" else 1] * 2, name=stack_name)

        window_size = (nn.shape[height_axis] // window_ratio, nn.shape[width_axis] // window_ratio)
        # window_size = (int(tf.math.ceil(nn.shape[height_axis] / window_ratio), int(tf.math.ceil(nn.shape[width_axis] / window_ratio))
        global_query = to_global_query(nn, window_ratio, num_head, activation=activation, name=stack_name + "q_global_")

        nn = nn if image_data_format() == "channels_last" else layers.Permute([2, 3, 1])(nn)  # channels_first -> channels_last
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            cur_global_query = None if block_id % 2 == 0 else global_query
            nn = gcvit_block(nn, window_size, num_head, cur_global_query, mlp_ratio, layer_scale, block_drop_rate, activation=activation, name=block_name)
            global_block_id += 1
        nn = nn if image_data_format() == "channels_last" else layers.Permute([3, 1, 2])(nn)  # channels_last -> channels_first
    nn = layer_norm(nn, name="pre_output_")

    nn = output_block(nn, num_classes=num_classes, drop_rate=dropout, classifier_activation=classifier_activation)
    model = models.Model(inputs, nn, name=model_name)
    add_pre_post_process(model, rescale_mode="torch")
    reload_model_weights(model, PRETRAINED_DICT, "gcvit", pretrained, MultiHeadRelativePositionalEmbedding)
    return model


@register_model
def GCViT_XXTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return GCViT(**locals(), model_name="gcvit_xx_tiny", **kwargs)


@register_model
def GCViT_XTiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 6, 5]
    return GCViT(**locals(), model_name="gcvit_x_tiny", **kwargs)


@register_model
def GCViT_Tiny(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 19, 5]
    return GCViT(**locals(), model_name="gcvit_tiny", **kwargs)


@register_model
def GCViT_Small(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 19, 5]
    num_heads = [3, 6, 12, 24]
    embed_dim = 96
    mlp_ratio = 2
    layer_scale = 1e-5
    return GCViT(**locals(), model_name="gcvit_small", **kwargs)


@register_model
def GCViT_Base(input_shape=(224, 224, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [3, 4, 19, 5]
    num_heads = [4, 8, 16, 32]
    embed_dim = 128
    mlp_ratio = 2
    layer_scale = 1e-5
    return GCViT(**locals(), model_name="gcvit_base", **kwargs)



from keras_cv_attention_models.efficientdet.efficientdet import (
    EfficientDet,
    EfficientDetD0,
    EfficientDetD1,
    EfficientDetD2,
    EfficientDetD3,
    EfficientDetD4,
    EfficientDetD5,
    EfficientDetD6,
    EfficientDetD7,
    EfficientDetD7X,
    EfficientDetLite0,
    EfficientDetLite1,
    EfficientDetLite2,
    EfficientDetLite3,
    EfficientDetLite3X,
    EfficientDetLite4,
)

__head_doc__ = """
Keras implementation of [google/automl/efficientdet](https://github.com/google/automl/tree/master/efficientdet).
Paper [Paper 1911.09070 EfficientDet: Scalable and Efficient Object Detection](https://arxiv.org/pdf/1911.09070.pdf).

Args:
  backbone: backbone model, could be any model with pyramid stage structure. {}
"""

__tail_doc__ = """  regression_len: bbox output len, typical value is 4, for yolov8 reg_max=16 -> regression_len = 16 * 4 == 64.
  anchors_mode: one of ["efficientdet", "anchor_free", "yolor"], controls which anchor to use.
      - efficientdet anchors default settings: use_object_scores=False, num_anchors=9, anchor_scale=4,
          aspect_ratios=[1, 2, 0.5], num_scales=3, grid_zero_start=False.
      - anchor_free default settings: use_object_scores=True, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=True.
      - yolor default settings: use_object_scores=True, num_anchors=3.
      - yolov8 default settings: use_object_scores=False, num_anchors=1, anchor_scale=1,
          aspect_ratios=[1], num_scales=1, grid_zero_start=False.
      Default "efficientdet".
  num_anchors: number of anchors for a single grid point, should be same with dataset used value.
      Default "auto" means: anchors_mode=="anchor_free" / "yolov8" -> 1, anchors_mode=="yolor" -> 3, else 9.
  use_object_scores: bollean value if model header output includes `object_scores`.
      Default "auto" means: True if anchors_mode=="anchor_free" or anchors_mode=="yolor", else False.
  num_classes: total output classes. `90` for EfficientDet pretrained, `80` for `tfds.coco`.
      Set `0` to disable `classifier` output.
  use_sep_conv: set `False` for using `Conv2D` instead of `SeparableConv2D`.
  activation: activation used in whole model, default `swish`. Default "swish".
  classifier_activation: The activation function to use for classifier output if `num_classes > 0`.
      Set `classifier_activation=None` to return the logits of the "top" layer. Default `sigmoid`.
  freeze_backbone: set `True` for `backbone.trainable = False`. Default `False`.
  pretrained: one of `None` (random initialization) or 'coco' (pre-training on COCO).
      Will try to download and load pre-trained model weights if not None. Default `coco`.
  pyramid_levels_min: anchors inititial parameter for model prediction, not affecting model architecture. Default `3`.
      pyramid_levels_max is calculated as `pyramid_levels_min + len(features_pick) + additional_features - 1`.
  anchor_scale: anchors inititial parameter for model prediction, not affecting model architecture.
      Default "auto" means: 1 if (anchors_mode=="anchor_free" or anchors_mode=="yolor"), else 4.
  rescale_mode: model precessing input, not for model structure. Defulat "torch".

Returns:
    A `keras.Model` instance.
"""

EfficientDet.__doc__ = __head_doc__.format("") + """  features_pick: specific `layer names` or `pyramid feature indexes` from backbone model.
        Default `[-3, -2, -1]` means using the last 3 pyramid feature output from backbone.
  additional_features: number of generated features from last backbone feature. `3` for EfficientDetD7X, `2` for others.
  fpn_depth: number of `bi_fpn` repeats.
  head_depth: depth for `bbox_regressor` and `classifier`.
  num_channels: channel filters for `bi_fpn`, `bbox_regressor` and `classifier`.
  use_weighted_sum: if use weighted sum in `bi_fpn`. `False` for `EfficientDetD6, D7, D7X`, `True` for others.
  model_name: string, model name.
""" + __tail_doc__ + """
Model architectures:
  | Model              | Params | FLOPs   | Input | COCO val AP | test AP |
  | ------------------ | ------ | ------- | ----- | ----------- | ------- |
  | EfficientDetD0     | 3.9M   | 2.55G   | 512   | 34.3        | 34.6    |
  | EfficientDetD1     | 6.6M   | 6.13G   | 640   | 40.2        | 40.5    |
  | EfficientDetD2     | 8.1M   | 11.03G  | 768   | 43.5        | 43.9    |
  | EfficientDetD3     | 12.0M  | 24.95G  | 896   | 46.8        | 47.2    |
  | EfficientDetD4     | 20.7M  | 55.29G  | 1024  | 49.3        | 49.7    |
  | EfficientDetD5     | 33.7M  | 135.62G | 1280  | 51.2        | 51.5    |
  | EfficientDetD6     | 51.9M  | 225.93G | 1280  | 52.1        | 52.6    |
  | EfficientDetD7     | 51.9M  | 325.34G | 1536  | 53.4        | 53.7    |
  | EfficientDetD7X    | 77.0M  | 410.87G | 1536  | 54.4        | 55.1    |
  | EfficientDetLite0  | 3.2M   | 0.98G   | 320   | 27.5        | 26.41   |
  | EfficientDetLite1  | 4.2M   | 1.97G   | 384   | 32.6        | 31.50   |
  | EfficientDetLite2  | 5.3M   | 3.38G   | 448   | 36.2        | 35.06   |
  | EfficientDetLite3  | 8.4M   | 7.50G   | 512   | 39.9        | 38.77   |
  | EfficientDetLite3X | 9.3M   | 14.01G  | 640   | 44.0        | 42.64   |
  | EfficientDetLite4  | 15.1M  | 20.20G  | 640   | 44.4        | 43.18   |
"""

__model_doc__ = __head_doc__ + __tail_doc__
EfficientDetD0.__doc__ = __model_doc__.format("Default None for EfficientNetV1B0.")
EfficientDetD1.__doc__ = __model_doc__.format("Default None for EfficientNetV1B1.")
EfficientDetD2.__doc__ = __model_doc__.format("Default None for EfficientNetV1B2.")
EfficientDetD3.__doc__ = __model_doc__.format("Default None for EfficientNetV1B3.")
EfficientDetD4.__doc__ = __model_doc__.format("Default None for EfficientNetV1B4.")
EfficientDetD5.__doc__ = __model_doc__.format("Default None for EfficientNetV1B5.")
EfficientDetD6.__doc__ = __model_doc__.format("Default None for EfficientNetV1B6.")
EfficientDetD7.__doc__ = __model_doc__.format("Default None for EfficientNetV1B6.")
EfficientDetD7X.__doc__ = __model_doc__.format("Default None for EfficientNetV1B7.")
EfficientDetLite0.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite0.")
EfficientDetLite1.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite1.")
EfficientDetLite2.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite2.")
EfficientDetLite3.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite3.")
EfficientDetLite3X.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite3X.")
EfficientDetLite4.__doc__ = __model_doc__.format("Default None for EfficientNetV1Lite4.")
